from msgbus import TK_SYS_NAK

from application.model import BenefitAppliers
from application.model.configuration import Configurations
from application.model.customexception import ErrorObtainingCouponInfo
from application.model.repository.benefit_applier import DefaultBenefitApplierRepository
from application.repository.pos import DBRepository
from application.repository.pos.utils import find_benefit_applier
from typing import Dict


class GetBenefitInfoInteractor(object):

    def __init__(self, configs, db_repository, benefit_appliers_repositories):
        # type: (Configurations, DBRepository, Dict[BenefitAppliers: DefaultBenefitApplierRepository]) -> None

        self.configs = configs
        self.logger = self.configs.logger
        self.db_repository = db_repository
        self.benefit_appliers_repositories = benefit_appliers_repositories

    def get_benefit_info(self, current_pos_id, voucher_id):
        # type: (int, str) -> str

        benefit_applier = find_benefit_applier(current_pos_id, voucher_id)

        self.logger.info("Finding benefit info. CurrentPosId: {}; VoucherId: {}".format(current_pos_id, voucher_id))

        response = self.benefit_appliers_repositories[benefit_applier].get_benefit_info(voucher_id)
        if response.token == TK_SYS_NAK:
            self.logger.info("Error finding benefit info. VoucherId: {}".format(voucher_id))
            raise ErrorObtainingCouponInfo(response.data)

        return response.data



import json
import logging

import sysactions

from xml.etree import cElementTree as eTree
from actions.eft import (
    save_cancel_receipt,
    confirm_finish_eft_transactions,
    set_eft_cache,
)

logger = logging.getLogger("EftActions")


@sysactions.action
def confirm_cancel_eft_mobile_transactions(
    pos_id,             # type: str
    data,               # type: str
    result_code=None,   # type: str
    message=None        # type: str
):
    # type: (...) -> None

    try:
        if not data or result_code != "0":
            error = "Error confirm cancel eft mobile transactions: {} | {} | {}".format(result_code, message, data)
            logger.error(error)
            if result_code not in ["-100", "-101"]:
                error_message = message if message else "$ERROR_CANCEL_ELECTRONIC_PAYMENTS_ORDER"
                sysactions.show_messagebox(pos_id, error_message)
            return

        model = sysactions.get_model(pos_id)
        order = sysactions.get_posot(model)

        data_cancel = json.loads(data)
        transaction_identifier = str(data_cancel.get("TransactionIdentifier", ""))
        identifier_len = len(transaction_identifier)
        order_id = int(transaction_identifier[:-3]) if identifier_len > 3 else ""
        tender_id = int(transaction_identifier[-3:]) if identifier_len > 3 else ""

        order_picture = eTree.XML(order.orderPicture(orderid=order_id)).find("Order")

        save_cancel_receipt(
            pos_id,
            order_picture,
            tender_id,
            data
        )

        from posactions import get_printer_receipt
        printer_receipt = get_printer_receipt()
        printer_receipt.print_tef_tender_receipt(
            pos_id=str(pos_id),
            tender_detail=data_cancel,
            fiscalization_type="APOS",
            print_customer_receipt=True,
        )

        set_eft_cache(
            pos_id=pos_id,
            payment_data=data_cancel,
            tender_id=tender_id,
            order_id=order_id,
        )

        logger.info("Success cancel transactions mobile: {0}".format(data))
    except Exception as _:
        logger.info("Error cancel transactions mobile: {0}".format(data))
    finally:
        confirm_finish_eft_transactions(pos_id, "1")



# Copyright (C) 2001-2006 Python Software Foundation
# Author: Barry Warsaw
# Contact: email-sig@python.org

"""email package exception classes."""
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from future.builtins import super


class MessageError(Exception):
    """Base class for errors in the email package."""


class MessageParseError(MessageError):
    """Base class for message parsing errors."""


class HeaderParseError(MessageParseError):
    """Error while parsing headers."""


class BoundaryError(MessageParseError):
    """Couldn't find terminating boundary."""


class MultipartConversionError(MessageError, TypeError):
    """Conversion to a multipart is prohibited."""


class CharsetError(MessageError):
    """An illegal charset was given."""


# These are parsing defects which the parser was able to work around.
class MessageDefect(ValueError):
    """Base class for a message defect."""

    def __init__(self, line=None):
        if line is not None:
            super().__init__(line)
        self.line = line

class NoBoundaryInMultipartDefect(MessageDefect):
    """A message claimed to be a multipart but had no boundary parameter."""

class StartBoundaryNotFoundDefect(MessageDefect):
    """The claimed start boundary was never found."""

class CloseBoundaryNotFoundDefect(MessageDefect):
    """A start boundary was found, but not the corresponding close boundary."""

class FirstHeaderLineIsContinuationDefect(MessageDefect):
    """A message had a continuation line as its first header line."""

class MisplacedEnvelopeHeaderDefect(MessageDefect):
    """A 'Unix-from' header was found in the middle of a header block."""

class MissingHeaderBodySeparatorDefect(MessageDefect):
    """Found line with no leading whitespace and no colon before blank line."""
# XXX: backward compatibility, just in case (it was never emitted).
MalformedHeaderDefect = MissingHeaderBodySeparatorDefect

class MultipartInvariantViolationDefect(MessageDefect):
    """A message claimed to be a multipart but no subparts were found."""

class InvalidMultipartContentTransferEncodingDefect(MessageDefect):
    """An invalid content transfer encoding was set on the multipart itself."""

class UndecodableBytesDefect(MessageDefect):
    """Header contained bytes that could not be decoded"""

class InvalidBase64PaddingDefect(MessageDefect):
    """base64 encoded sequence had an incorrect length"""

class InvalidBase64CharactersDefect(MessageDefect):
    """base64 encoded sequence had characters not in base64 alphabet"""

# These errors are specific to header parsing.

class HeaderDefect(MessageDefect):
    """Base class for a header defect."""

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)

class InvalidHeaderDefect(HeaderDefect):
    """Header is not valid, message gives details."""

class HeaderMissingRequiredValue(HeaderDefect):
    """A header that must have a value had none"""

class NonPrintableDefect(HeaderDefect):
    """ASCII characters outside the ascii-printable range found"""

    def __init__(self, non_printables):
        super().__init__(non_printables)
        self.non_printables = non_printables

    def __str__(self):
        return ("the following ASCII non-printables found in header: "
            "{}".format(self.non_printables))

class ObsoleteHeaderDefect(HeaderDefect):
    """Header uses syntax declared obsolete by RFC 5322"""

class NonASCIILocalPartDefect(HeaderDefect):
    """local_part contains non-ASCII characters"""
    # This defect only occurs during unicode parsing, not when
    # parsing messages decoded from binary.



import os
import sys

from application.configurator import Configurator
from application.repository import MessageBusRepository
from eventhandlerbuilder import AuditEventConsumerEventHandler
from mbcontextmessagehandler import MbContextMessageBus

from helper import import_pydevd

reload(sys)
sys.setdefaultencoding("utf-8")


def main():
    # type: () -> None

    loader_path = os.environ["LOADERCFG"]
    import_pydevd(loader_path, 9136)

    configs = Configurator(loader_path)
    configs.configure_service_name()
    configs.configure_logger()

    configs.logger.info("Starting Pump Component...")

    event_handler = AuditEventConsumerEventHandler(configs)
    message_bus = MbContextMessageBus(event_handler.mbcontext)

    repository = MessageBusRepository(message_bus)
    configs.read_configurations(repository)

    event_handler.set_repository(repository)
    event_handler.build_interactors()
    event_handler.start()

    configs.logger.info("Finishing Pump Component...")


if __name__ == "__main__":
    main()



# encoding: utf-8
from datetime import datetime

from old_helper import BaseRepository
from msgbus import MBEasyContext
from persistence import Connection
from typing import List, Dict, Optional

from ..model import Order, OrderTender


class OrderRepository(BaseRepository):
    TypeRealDate = "RealDate"
    TypeBusinessPeriod = "BusinessPeriod"
    TypeSessionId = "SessionId"

    def __init__(self, mbcontext, pos_list):
        # type: (MBEasyContext, List[int]) -> None
        super(OrderRepository, self).__init__(mbcontext)
        self.pos_list = pos_list

    def get_paid_orders_by_real_date(self, initial_date, end_date, operator_id, report_pos):
        # type: (datetime, datetime, unicode, unicode) -> List[Order]
        return self._get_paid_orders(self.TypeRealDate, initial_date, end_date, operator_id, None, report_pos)

    def get_paid_orders_by_business_period(self, initial_date, end_date, operator_id, report_pos):
        # type: (datetime, datetime, unicode, unicode) -> List[Order]
        return self._get_paid_orders(self.TypeBusinessPeriod, initial_date, end_date, operator_id, None, report_pos)

    def get_paid_orders_by_session_id(self, session_id):
        # type: (unicode) -> List[Order]
        return self._get_paid_orders(self.TypeSessionId, None, None, None, session_id, None)

    def _get_paid_orders(self, report_type, initial_date, end_date, operator_id, session_id, report_pos):
        # type: (unicode, Optional[datetime], Optional[datetime], Optional[unicode], Optional[unicode], Optional[unicode]) -> List[Order]
        def inner_func(conn):
            # type: (Connection) -> List[Order]
            try:
                if report_type == self.TypeRealDate:
                    query = self._PaidOrdersByRealDateQuery
                    date_format = "%Y-%m-%d"
                    query = query.format(initial_date.strftime(date_format), end_date.strftime(date_format))
                elif report_type == self.TypeBusinessPeriod:
                    query = self._PaidOrderByBusinessPeriod
                    date_format = "%Y%m%d"
                    query = query.format(initial_date.strftime(date_format), end_date.strftime(date_format))
                else:
                    query = self._PaidOrderBySessionId
                    date_format = "%Y%m%d"
                    query = query.format(session_id)

                if report_type == self.TypeRealDate or self.TypeBusinessPeriod and operator_id is not None:
                    query += " and o.SessionId like '%user={0},%'".format(operator_id)

                all_orders = {}  # type: Dict[int, Order]
                order_with_payments = [(int(x.get_entry(0)),
                                        int(x.get_entry(1)),
                                        datetime.strptime(x.get_entry(2), date_format),
                                        float(x.get_entry(3)),
                                        int(x.get_entry(4)),
                                        float(x.get_entry(5))) for x in conn.select(query)]

                for order_tuple in order_with_payments:
                    order_id = order_tuple[0]
                    order_state = order_tuple[1]
                    order_date = order_tuple[2]
                    order_total = order_tuple[3]

                    if order_id in all_orders:
                        order = all_orders[order_id]
                    else:
                        order = Order(order_id, order_date, order_state, order_total, [])
                        all_orders[order_id] = order

                    tender_type = order_tuple[4]
                    tender_value = order_tuple[5]

                    order_tender = OrderTender(tender_type, tender_value)
                    order.tenders.append(order_tender)

                return list(all_orders.values())
            except Exception as _:
                return list({"error": "error"}.values())

        report_pos_list = self.pos_list if report_pos is None else (report_pos,)
        return self.execute_parallel_in_all_databases_returning_flat_list(inner_func, report_pos_list)

    _PaidOrdersByRealDateQuery = \
"""select o.OrderId, o.StateId, od.OrderDate, o.TotalGross, ot.TenderId, ot.TenderAmount - coalesce(ot.ChangeAmount, 0)
from Orders o
inner join OrderTender ot
on o.OrderId = ot.OrderId
inner join 
(
	-- Busca as Orders do dia que estamos interessadosh
	select OrderId, strftime('%Y-%m-%d', Value, 'localtime') OrderDate
	from OrderCustomProperties
	where key = 'FISCALIZATION_DATE' and strftime('%Y-%m-%d', Value, 'localtime') >= '{0}' and strftime('%Y-%m-%d', Value, 'localtime') <= '{1}'
) od
on o.OrderId = od.OrderId
where o.StateId = 5"""

    _PaidOrderByBusinessPeriod = \
"""select o.OrderId, o.StateId, o.BusinessPeriod, o.TotalGross, ot.TenderId, ot.TenderAmount - coalesce(ot.ChangeAmount, 0)
from Orders o
inner join OrderTender ot
on o.OrderId = ot.OrderId
where o.BusinessPeriod >= '{0}' and o.BusinessPeriod <= '{1}' and o.StateId = 5"""

    _PaidOrderBySessionId = \
"""select o.OrderId, o.StateId, o.BusinessPeriod, o.TotalGross, ot.TenderId, ot.TenderAmount - coalesce(ot.ChangeAmount, 0)
from Orders o
inner join OrderTender ot
on o.OrderId = ot.OrderId
where o.SessionId = '{0}' and o.StateId = 5"""



# -*- coding: utf-8 -*-
import inspect

from xml.etree import cElementTree as eTree
from functools import wraps
from datetime import datetime

from sysactions import get_model, StopAction, show_messagebox, get_podtype, get_line_product_name, get_posot, get_custom
from systools import sys_log_exception

from actions.config import get_function_authorization_level_from_config
from actions.util import get_user_level, get_authorization
from actions.pos import pos_need_to_be_blocked

configurations = None


def initialize_manager_authorizations(config):
    global configurations
    configurations = config


def user_authorized_to_execute_action(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        user_data = None
        pos_id = _get_pos_id(args, kwargs)
        if not pos_id:
            return False

        try:
            function_level = get_function_authorization_level_from_config(configurations, f.func_name)
            model = get_model(pos_id)
            operator = model.find("./Operator")
            if operator is not None:
                user_id = operator.get("id")
                if get_user_level(user_id) <= function_level:
                    status, user_data = get_authorization(pos_id, function_level, model)

                    if status:
                        _save_authentication_info_on_order(f.func_name, model, pos_id)
                    else:
                        return StopAction()

        except BaseException as ex: # noqa
            error_message = "$GET_USER_AUTHORIZATION_ERROR_" + str(type(ex).__name__).upper()
            sys_log_exception("Error validating user authorization")
            show_messagebox(pos_id, message=error_message, icon="error")
            return False

        if 'auth_data' in inspect.getargspec(f).args:
            kwargs['auth_data'] = user_data

        return f(*args, **kwargs)

    return decorated_function


def block_action_if_pos_is_blocked(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        pos_id = _get_pos_id(args, kwargs)
        if not pos_id:
            return False

        try:
            model = get_model(pos_id)
            pod_type = get_podtype(model)
            if pod_type != 'TT' and pos_need_to_be_blocked(pos_id):
                from msgbusboundary import MsgBusTableService as tableService
                from table_actions import deselect_table
                selected_table = tableService().get_selected_table(pos_id)
                if selected_table is None:
                    deselect_table(pos_id)

                show_messagebox(pos_id, message="$POS_BLOCKED", icon="error")
                return StopAction()
        except BaseException as _:  # noqa
            return False

        return f(*args, **kwargs)

    return decorated_function


def block_action_if_order_items_not_resolved(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        from posactions import list_open_options

        pos_id = _get_pos_id(args, kwargs)
        if not pos_id:
            return False

        model = get_model(pos_id)
        order = model.find("CurrentOrder/Order")

        sale_lines = order.findall("SaleLine")
        active_sale_lines = _get_active_order_lines(sale_lines)
        is_remote_order = order.find("CustomOrderProperties/OrderProperty[@key='REMOTE_ORDER_ID']")
        if not active_sale_lines and is_remote_order is None:
            show_messagebox(pos_id, "$HAS_NO_ITEMS_IN_ORDER", title="")
            return False

        option = list_open_options(order)
        if option is not None:
            prod_name = get_line_product_name(model, int(option.get("lineNumber")))
            show_messagebox(pos_id, message="$NEED_TO_RESOLVE_OPTION|%s" % (prod_name.encode("UTF-8")), icon="error")
            return False

        return f(*args, **kwargs)

    return decorated_function


def _get_pos_id(args, kwargs):
    pos_id = None
    if len(args) > 0:
        pos_id = args[0]
    elif 'pos_id' in kwargs:
        pos_id = kwargs['pos_id']
    elif 'posid' in kwargs:
        pos_id = kwargs['posid']
    else:
        sys_log_exception("PosId not found to authorize")

    return pos_id


def _get_order_deleted_lines(sale_lines):
    return map(lambda x: x.get("lineNumber"), filter(lambda x: x.get("level") == "0" and x.get("qty") == "0", sale_lines))


def _get_active_order_lines(sale_lines):
    deleted_line_numbers = _get_order_deleted_lines(sale_lines)
    return filter(lambda x: x.get("lineNumber") not in deleted_line_numbers, sale_lines)


def _save_authentication_info_on_order(
        executed_function,      # type: str
        model,                  # type: eTree.Element
        pos_id                  # type: int
):
    # type: (...) -> None

    if executed_function is not None:
        if not model:
            model = get_model(pos_id)

        pos_ot = get_posot(model)
        user_id = get_custom(model, 'Last Manager ID')
        order = model.find("CurrentOrder/Order")
        order_id = order.get("orderId") if order else None

        if order_id:
            property_name = "{}: {}".format(executed_function, datetime.now().strftime("%m/%d/%Y %H:%M:%S"))
            pos_ot.setOrderCustomProperty(property_name, user_id, order_id)



from helper import ExtendedEnum


class AddBenefitValidationErrorCode(ExtendedEnum):
    NDISCOUNT_ERROR_USER_MIN_LEVEL = "The user applying the benefit does not have the level required by the benefit"



import json

from application.interactor.apply.apply_benefit import ApplyBenefitInteractor
from application.interactor.apply.apply_benefit.dto.request import ApplyBenefitRequest
from application.interactor.apply.dto import ApplyContext
from .._BaseApplyProcessor import BaseApplyProcessor


class ApplyBenefitProcessor(BaseApplyProcessor):

    def __init__(
        self,        # type: ApplyBenefitProcessor
        interactor,  # type: ApplyBenefitInteractor
    ):
        # type: (...) -> None

        super(ApplyBenefitProcessor, self).__init__()
        self.interactor = interactor

    def parse_data(
        self,        # type: ApplyBenefitProcessor
        data,        # type: bytes
        event_type,  # type: str
    ):
        # type: (...) -> (int, int, str, int)

        parsed_data = json.loads(data)
        pos_id = int(parsed_data.get("posId"))
        originator_pos_id = int(parsed_data.get("originatorPosId") or pos_id)
        benefit = parsed_data["benefit"]
        user_auth_level = parsed_data.get("user_auth_level")
        if user_auth_level is None:
            user_auth_level = -1
        applier = parsed_data["applier"]

        return ApplyBenefitRequest(
            benefit=benefit,
            context=ApplyContext(
                pos_id=pos_id,
                originator_pos_id=originator_pos_id,
                user_auth_level=user_auth_level,
                applier=applier
            )
        )

    def call_business(
            self,     # type: ApplyBenefitProcessor
            request,  # type: ApplyBenefitRequest
    ):
        # type: (...) -> None

        self.interactor.apply_benefit(apply_benefit_request=request)



from _OrderRetrieverService import OrderRetrieverService
from _OrderRetrieverException import OrderRetrieverException
from _OrderDataFormatter import OrderDataFormatter


__all__ = ["OrderRetrieverService", "OrderRetrieverException", "OrderDataFormatter"]



import json
import sys
from logging import Logger

from messagebus import (
    Event,
    MessageBus,
)
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Union
)
from xml.etree import cElementTree as eTree

from posot import OrderTaker
from sysactions import get_tender_types

from application.model.benefit import (
    Action,
    Benefit,
    RequiredItemOnSale,
    OrderBenefitInfo,
    ActionResult,
    Conditions,
)
from application.model.conditions import Restrictions
from application.model.exceptions import (
    InvalidTenderIdException,
    InvalidAmountException,
    NotEffectiveDiscountException,
    InvalidQuantityException,
    AddItemsOnTotalException
)
from application.model.order import Order, SaleLine
from application.model.transactions import AddBenefitToApplyUnitOfWork
from application.util import (
    CompositionType,
    add_to_dict_value,
    get_item_values_line_ids,
    get_line_id_from_sale_line,
    get_remaining_amount_discount,
    have_action_by_items_or_groups,
    is_invalid_sale_line_or_remaining_quantity,
    subtract_quantity,
    round_value,
)

FAKE_DISCOUNT_ID = "999"


class DiscountService(object):
    def __init__(
        self,
        pos_ot,             # type: OrderTaker
        product_groups,     # type: Dict
        message_bus,        # type: MessageBus
        logger,             # type: Logger
    ):
        # type: (...) -> None

        self.pos_ot = pos_ot
        self.product_groups = product_groups
        self.message_bus = message_bus
        self.logger = logger

    def apply_discount_amount_on_total(
        self,                    # type: DiscountService
        action,                  # type: Action
        order,                   # type: Order
        restriction,             # type: Restrictions
        available_discount_ids,  # type: set
        benefit,                 # type: Benefit
    ):
        # type: (...) -> ActionResult

        return self._apply_discount_on_total(
            action=action,
            order=order,
            restriction=restriction,
            available_discount_ids=available_discount_ids,
            benefit=benefit,
        )

    def apply_discount_percentage_on_total(
        self,
        action,                     # type: Action
        order,                      # type: Order
        restriction,                # type: Restrictions
        available_discount_ids,     # type: set
        benefit,                    # type: Benefit
    ):
        # type: (...) -> ActionResult

        return self._apply_discount_on_total(
            action=action,
            order=order,
            restriction=restriction,
            available_discount_ids=available_discount_ids,
            benefit=benefit,
            use_percentage=True,
        )

    def apply_discount_amount_by_groups(self, action, order, restriction, discount_ids, apply_on_total):
        # type: (Action, Order, Restrictions, set, bool) -> ActionResult

        return self._apply_discount_by_items(action, order, restriction, discount_ids, False, apply_on_total, True)

    def apply_discount_percentage_by_groups(self, action, order, restriction, discount_ids, apply_on_total):
        # type: (Action, Order, Restrictions, set, bool) -> ActionResult

        return self._apply_discount_by_items(action, order, restriction, discount_ids, True, apply_on_total, True)

    def apply_discount_amount_by_items(self, action, order, restriction, discount_ids, apply_on_total):
        # type: (Action, Order, Restrictions, set, bool) -> ActionResult

        return self._apply_discount_by_items(action, order, restriction, discount_ids, False, apply_on_total)

    def apply_discount_percentage_by_items(self, action, order, restriction, discount_ids, apply_on_total):
        # type: (Action, Order, Restrictions, set, bool) -> ActionResult

        return self._apply_discount_by_items(action, order, restriction, discount_ids, True, apply_on_total)

    def apply_discount_amount_on_total_by_tender(self, action, order, restriction, by_items=False):
        # type: (Action, Order, Restrictions, bool) -> ActionResult

        return self.apply_discount_on_total_by_tender(action, order, restriction, by_items=by_items)

    def apply_discount_percentage_on_total_by_tender(self, action, order, restriction):
        # type: (Action, Order, Restrictions) -> ActionResult

        return self.apply_discount_on_total_by_tender(action, order, restriction, True)

    def apply_items_to_add(self, action, burnt_items, uow, restriction, on_total=False):
        # type: (Action, Dict, AddBenefitToApplyUnitOfWork, Restrictions, bool) -> ActionResult

        items_to_burn = {}
        benefit_applied = True

        for entry in action.value:
            part_code = entry.get("item")
            quantity = entry.get("quantity")
            free = bool(entry.get("free"))
            price = entry.get("price", None)
            tender_id = entry.get("tenderId", None)

            quantity = self._get_quantity_to_apply(burnt_items, quantity, uow)
            items_to_burn.update(burnt_items)
            for _ in range(0, quantity):
                action_burnt_items = uow.add_sale_line(
                    part_code=part_code,
                    is_free=free,
                    price=price,
                    tender_id=tender_id,
                    burnt_items=items_to_burn,
                    restriction=restriction,
                    on_total=on_total,
                    action=action,
                )
                items_to_burn.update(action_burnt_items)
                benefit_applied = benefit_applied and not free

        return ActionResult([], items_to_burn, [], {}, benefit_applied, not on_total)

    def apply_tender_on_added_items(self, action, order, benefit_order_info):
        # type: (Action, Order, OrderBenefitInfo) -> ActionResult

        if benefit_order_info is None:
            raise AddItemsOnTotalException()

        order_tender_ids = []

        if not benefit_order_info.promotion_tender:
            return ActionResult([], {}, order_tender_ids, {}, True, True)

        for promotion_tender in benefit_order_info.promotion_tender:
            if action.order != promotion_tender.get("action_id"):
                continue

            tender_amount = promotion_tender.get("amount", 0)
            if tender_amount > 0:
                promotion_tender_id = promotion_tender.get("tender_id")
                if promotion_tender_id is None:
                    raise InvalidTenderIdException()

                tender_id = int(promotion_tender_id)
                pos_id = order.pos_id
                order_tender_id = self._do_order_tender(pos_id, tender_id, tender_amount)
                if order_tender_id:
                    order_tender_ids.append(order_tender_id)

        return ActionResult([], {}, order_tender_ids, {}, True, False)

    def _show_message(
        self,
        pos_id,                 # type: str
        message,                # type: str
        title="$INFORMATION",   # type: str
    ):
        # type: (...) -> None

        if not message:
            return
        data = "{}\\0{}\\0{}\\0{}".format(pos_id, message, title, True)
        self.message_bus.publish_event(Event("SHOW_MESSAGE", "NDISCOUNT", data))

    def _apply_discount_on_total(
        self,
        action,                     # type: Action
        order,                      # type: Order
        restriction,                # type: Restrictions
        available_discount_ids,     # type: set
        benefit,                    # type: Benefit
        use_percentage=False,       # type: bool
    ):
        # type: (...) -> ActionResult

        amount = self._get_amount(
            amount=action.value,
            order=order,
            restriction=restriction,
            seat=action.seat,
            use_percentage=use_percentage,
            persistent_value=action.persistent_value,
        )

        max_value_condition = sys.maxint
        if benefit.conditions:
            if benefit.conditions.discount_max_value is not None:
                max_value_condition = benefit.conditions.discount_max_value - order.discount_amount
            if action.max_value:
                max_value_condition = action.max_value - order.discount_amount

        to_apply_amount = min(float(amount), max_value_condition)
        inserted_value = amount if use_percentage else action.value
        self._validate_discount_amount(
            pos_id=order.pos_id,
            inserted_value=inserted_value,
            max_value_condition=max_value_condition,
            to_apply_amount=to_apply_amount,
            percentage_rate=action.value if use_percentage else None,
        )

        if to_apply_amount > 0:
            to_tender_amount = 0
            total_gross = order.total_gross
            if benefit.tip_included:
                total_gross += order.tip

            if round_value(to_apply_amount - total_gross) >= 0:
                to_tender_amount = self._calculate_to_tender_amount(benefit, order)
                to_apply_amount = to_apply_amount - to_tender_amount

            discount_id = available_discount_ids.pop()
            self.pos_ot.applyDiscount(
                discountid=discount_id,
                discountamt=to_apply_amount,
                forcebyitem=1,
            )
            discount_by_line_number = self._split_discount_by_sale_line(
                order=order,
                restriction=restriction,
                discount_id=str(discount_id),
                discount_amount=to_apply_amount,
            )

            order_tender_ids = []
            if to_tender_amount > 0:
                tender_id = self._get_tender_id(
                    action=action,
                )
                if tender_id is not None:
                    order_tender_id = self._do_order_tender(order.pos_id, tender_id, to_tender_amount)
                    if order_tender_id is not None:
                        order_tender_ids.append(order_tender_id)

            return ActionResult([discount_id], dict(), order_tender_ids, discount_by_line_number, True, False)

        return ActionResult([], {}, [], {}, False, False)

    def _validate_discount_amount(
        self,                       # type: DiscountService
        pos_id,                     # type: str
        inserted_value,             # type: float
        max_value_condition,        # type: float
        to_apply_amount,            # type: float
        percentage_rate,            # type: Optional[float]
    ):
        # type: (...) -> None

        message = None
        if percentage_rate is not None:
            if inserted_value > to_apply_amount:
                message = '$NDISCOUNT_LIMITED_PERCENTAGE_DISCOUNT|{0:.2f}|{1:.2f}|{2:.2f}'.format(
                    percentage_rate,
                    max_value_condition,
                    max_value_condition,
                )
        else:
            if inserted_value > to_apply_amount:
                message = '$NDISCOUNT_LIMITED_AMOUNT_DISCOUNT|{0:.2f}'.format(
                    to_apply_amount
                )

        if message:
            self._show_message(
                pos_id=pos_id,
                message=message,
            )

    def _calculate_to_tender_amount(self, benefit, order):
        total_tender = sum([x for x in (order.tenders or {}).itervalues()])
        if round_value(total_tender) > 0.00:
            return 0.00

        if self._is_table_order(order) and benefit.tip_included and order.tip > 0:
            return order.tip

        return 0.01

    @staticmethod
    def _is_table_order(
        order,  # type: Order
    ):
        return "SERVICE_ID" in order.custom_properties

    def _get_discounts_in_order_picture_by_sale_line(
        self,
        current_discount_id,     # type: str
        order_id,                # type: int
        order_picture_xml=None,  # type: str
    ):
        # type: (...) -> Dict

        if not order_picture_xml:
            order_picture_xml = self.pos_ot.orderPicture(
                orderid=order_id,
            )
            order_picture = eTree.XML(order_picture_xml).find("Order")
        else:
            order_picture = eTree.XML(order_picture_xml)

        sale_lines = order_picture.findall("SaleLine")
        discount_by_line_number = {}
        for sale_line in sale_lines:
            if not sale_line.get("discountsApplied"):
                continue
            discounts_applied = sale_line.get("discountsApplied").split(',')
            for discount_applied in discounts_applied:
                discount_id, discount_amount = discount_applied.split(':')
                if current_discount_id != int(discount_id):
                    continue

                if discount_id not in discount_by_line_number:
                    discount_by_line_number.update({
                        discount_id: {},
                    })
                line_id = get_line_id_from_sale_line(
                    sale_line=sale_line,
                    from_xml=True,
                )

                discount_line = float(discount_amount) + discount_by_line_number[discount_id].get(line_id, 0.0)
                discount_by_line_number[discount_id].update({
                    line_id: discount_line,
                })

        return discount_by_line_number

    def _apply_discount_by_items(
        self,
        action,                 # type: Action
        order,                  # type: Order
        restriction,            # type: Restrictions
        usable_discount_ids,    # type: set
        use_percentage,         # type: bool
        apply_on_total,         # type: bool
        by_groups=False         # type: bool
    ):
        # type: (...) -> ActionResult

        applied_discounts = []
        discount_by_group = None
        remaining_group_qty = None
        if by_groups:
            action_quantity = int(action.value[0].get("quantity", "1"))
            remaining_group_qty = (action.times_to_apply or 1) * action_quantity
            items_to_apply_discount, discount_by_group = self._get_items_by_groups_to_apply_discount(
                action=action,
                use_percentage=use_percentage,
            )
        else:
            items_to_apply_discount = self._get_items_to_apply_discount(action)

        discount_amount_on_total = 0
        total_discount = 0
        item_values = {}
        discount_by_line_number = {}
        entry_sale_lines = self._filter_sale_lines_by_seat(
            action=action,
            order=order,
        )
        benefits_to_remove = []
        for sale_line in entry_sale_lines:
            is_invalid = is_invalid_sale_line_or_remaining_quantity(
                restriction=restriction,
                sale_line=sale_line,
                remaining_quantity=remaining_group_qty,
            )
            if is_invalid:
                continue

            current_discount = items_to_apply_discount.get(int(sale_line.item_id))
            if current_discount is None:
                continue

            message = "Applying discount by items: {} / {} / {}"
            self.logger.info(message.format(current_discount, sale_line.key, sale_line.qty))
            remaining_sale_line_qty = sale_line.qty
            amount = 0

            while remaining_sale_line_qty > 0 and current_discount is not None:
                selected_discount = current_discount
                group = selected_discount["group"] if discount_by_group else ""
                if discount_by_group:
                    quantity_to_be_removed = min(remaining_sale_line_qty, discount_by_group[group]["quantity"])
                else:
                    quantity_to_be_removed = min(remaining_sale_line_qty, selected_discount["quantity"])

                if use_percentage:
                    total_value_to_remove = self._get_item_price(
                        order=order,
                        sale_line=sale_line,
                        quantity=quantity_to_be_removed,
                        current_price=False,
                    )
                    if not apply_on_total:
                        max_value_to_apply = self._get_item_price(
                            order=order,
                            sale_line=sale_line,
                            quantity=quantity_to_be_removed,
                        )
                        total_value_to_remove = min(total_value_to_remove, max_value_to_apply)
                else:
                    total_value_to_remove = self._get_item_price(
                        order=order,
                        sale_line=sale_line,
                        quantity=quantity_to_be_removed,
                    )

                if not discount_by_group:
                    remaining_sale_line_qty -= quantity_to_be_removed
                    selected_discount["quantity"] -= quantity_to_be_removed
                    remaining_group_qty = subtract_quantity(
                        value=remaining_group_qty,
                        value_to_subtract=quantity_to_be_removed,
                    )

                if use_percentage:
                    amount_to_add = total_value_to_remove * (selected_discount["amount"] / 100.0)
                else:
                    if discount_by_group:
                        amount_to_add = discount_by_group[group]["amount"]
                    else:
                        amount_to_add = quantity_to_be_removed * selected_discount["amount"]

                amount += min(amount_to_add, total_value_to_remove)
                if discount_by_group and amount > 0:
                    discount_by_group[group]["amount"] -= min(amount_to_add, total_value_to_remove)
                    discount_by_group[group]["quantity"] -= quantity_to_be_removed
                    remaining_sale_line_qty -= quantity_to_be_removed
                    selected_discount["quantity"] -= quantity_to_be_removed
                    remaining_group_qty = subtract_quantity(
                        value=remaining_group_qty,
                        value_to_subtract=quantity_to_be_removed,
                    )

                if discount_by_group and group in discount_by_group:
                    if discount_by_group[group]["quantity"] == 0 or discount_by_group[group]["amount"] == 0:
                        break

                if remaining_group_qty == 0 or selected_discount["quantity"] == 0 or selected_discount["amount"] == 0:
                    break

            max_discount_amount = self._get_order_max_discount_amount(
                order=order,
                restriction=restriction,
                seat=action.seat,
            )
            persistent_amount = get_remaining_amount_discount(
                max_value_to_discount=action.persistent_value,
                actual_value=total_discount,
            )
            amount = min(float(amount), max_discount_amount, persistent_amount)
            message = "Applying discount by items: sale line amount - {} / {} / {}"
            self.logger.info(message.format(current_discount, sale_line.key, amount))
            if amount > 0:
                line_id = get_line_id_from_sale_line(sale_line=sale_line)
                benefits_to_remove = self._add_id_if_needs_to_remove_previous_benefits(
                    order=order,
                    line_id=line_id,
                    benefits_ids_to_remove=benefits_to_remove,
                    apply_on_total=apply_on_total,
                )
                if benefits_to_remove:
                    continue

                add_to_dict_value(
                    object_dict=item_values,
                    dict_key=line_id,
                    value=amount,
                )

                if apply_on_total:
                    discount_amount_on_total += amount
                else:
                    discount_id = usable_discount_ids.pop()
                    self._apply_discount(
                        amount=amount,
                        discount_id=discount_id,
                        line_number=sale_line.line_number,
                    )

                    applied_discounts.append(discount_id)
                    discount_by_line_number[discount_id] = item_values

                    item_values = {}

                total_discount += amount

        order_tender_ids = []
        if discount_amount_on_total > 0 and not benefits_to_remove:
            max_discount_amount = self._get_order_max_discount_amount(
                order=order,
                restriction=restriction,
                seat=action.seat,
            )
            persistent_amount = get_remaining_amount_discount(max_value_to_discount=action.persistent_value)
            discount_amount_on_total = min(float(discount_amount_on_total), max_discount_amount, persistent_amount)

            to_tender_amount = 0
            total_gross = order.total_gross
            if restriction.benefit.tip_included:
                total_gross += order.tip

            if round_value(discount_amount_on_total - total_gross) >= 0:
                to_tender_amount = self._calculate_to_tender_amount(restriction.benefit, order)
                discount_amount_on_total = discount_amount_on_total - to_tender_amount

            discount_id = usable_discount_ids.pop()
            self._apply_discount(
                amount=discount_amount_on_total,
                discount_id=discount_id,
                line_number=None,
            )
            discount_by_line_number[discount_id] = item_values
            applied_discounts.append(discount_id)

            if to_tender_amount > 0:
                tender_id = self._get_tender_id(
                    action=action,
                )
                if tender_id is not None:
                    order_tender_id = self._do_order_tender(order.pos_id, tender_id, to_tender_amount)
                    if order_tender_id is not None:
                        order_tender_ids.append(order_tender_id)

        self.logger.info("Applying discount by items: final total discount - {}".format(total_discount))
        if total_discount > 0 or benefits_to_remove:
            return ActionResult(
                discounts_id=applied_discounts,
                items_to_burn={},
                order_tender_ids=order_tender_ids,
                discount_item_values=discount_by_line_number,
                benefit_applied=True,
                pending_to_apply=False,
                benefit_ids_to_remove=benefits_to_remove,
            )
        else:
            self.logger.warn("Non effective discount. Total discount: {}".format(total_discount))
            raise NotEffectiveDiscountException()

    def apply_discount_on_total_by_tender(self, action, order, restriction, use_percentage=False, by_items=False):
        # type: (Action, Order, Restrictions, bool, bool) -> ActionResult

        order_tender_ids = []
        for entry in action.value:
            tender_id = int(entry.get("tenderId", -1))
            if not self._validate_tender(tender_id):
                raise InvalidTenderIdException()

            if by_items:
                self._apply_payments_by_items(
                    entry=entry,
                    order=order,
                    order_tender_ids=order_tender_ids,
                    restriction=restriction,
                    tender_id=tender_id,
                    persistent_value=action.persistent_value,
                )
            else:
                self._apply_payment_on_order(
                    entry=entry,
                    order=order,
                    order_tender_ids=order_tender_ids,
                    restriction=restriction,
                    tender_id=tender_id,
                    use_percentage=use_percentage,
                    persistent_value=action.persistent_value,
                )

        return ActionResult([], {}, order_tender_ids, {}, True, False)

    def _apply_payments_by_items(
        self,
        entry,              # type: Dict
        order,              # type: Order
        order_tender_ids,   # type: List[int]
        restriction,        # type: Restrictions
        tender_id,          # type: int
        persistent_value,   # type: float
    ):
        # type: (...) -> None

        seat = entry.get("seat", None)
        sale_lines = self._get_sale_lines_with_seat(
            order=order,
            seat=seat,
        )

        quantity = self._get_quantity_to_apply_discount(
            sale_lines=sale_lines,
            conditions=restriction.benefit.conditions,
        )
        if not quantity:
            return

        total_tender_amount = 0
        for sale_line in order.sale_lines:
            item_id = sale_line.item_id
            if restriction.validate(sale_line) and item_id == str(entry["item"]):
                available_quantity = min(sale_line.qty, quantity)
                amount = self._get_item_price(
                    order=order,
                    sale_line=sale_line,
                    quantity=available_quantity,
                )
                persistent_amount = get_remaining_amount_discount(
                    max_value_to_discount=persistent_value,
                    actual_value=total_tender_amount,
                )
                amount = min(amount, persistent_amount)
                order_tender_id = self._do_order_tender(order.pos_id, tender_id, amount) if amount else False
                if order_tender_id:
                    order_tender_ids.append(order_tender_id)
                    line_number = sale_line.line_number
                    self.pos_ot.setOrderCustomProperty(
                            "discountPayment", order_tender_id, "", line_number, "1", "0", item_id
                    )
                    total_tender_amount += amount

                quantity -= available_quantity
                if quantity == 0:
                    break

    def _apply_payment_on_order(
        self,
        entry,              # type: Dict
        order,              # type: Order
        order_tender_ids,   # type: List[int]
        restriction,        # type: Restrictions
        tender_id,          # type: int
        use_percentage,     # type: bool
        persistent_value,   # type: Optional[float]
    ):
        # type: (...) -> None

        value = float(entry.get("percentage", 0)) if use_percentage else float(entry.get("amount", 0))

        amount = self._get_amount(
            amount=value,
            order=order,
            restriction=restriction,
            seat=entry.get("seat"),
            use_percentage=use_percentage,
            persistent_value=persistent_value,
        )

        if not amount:
            raise NotEffectiveDiscountException()

        order_tender_id = self._do_order_tender(order.pos_id, tender_id, amount)
        if order_tender_id:
            order_tender_ids.append(order_tender_id)

    def _apply_discount(
        self,
        amount,             # type: float
        discount_id,        # type: int
        line_number=None,   # type: Optional[int]
    ):
        # type: (...) -> str

        if line_number is None:
            order_picture_xml = self.pos_ot.applyDiscount(discount_id, amount, forcebyitem=1)
        else:
            order_picture_xml = self.pos_ot.applyDiscount(
                discountid=discount_id,
                discountamt=amount,
                linenumber=line_number,
            )

        return order_picture_xml

    def _get_total_amount_with_restrictions(
        self,
        order,          # type: Order
        restriction,    # type: Restrictions
        seat,           # type: int
    ):
        # type: (...) -> float

        if seat is not None:
            order.get_table_info()

        amount = 0
        discount_line_values = self._split_discount_by_sale_line(
            order=order,
            restriction=restriction,
            discount_id=FAKE_DISCOUNT_ID,
            discount_amount=sys.maxint,
            seat=seat,
        )
        total_gross = order.total_gross
        if not discount_line_values:
            return total_gross - order.discount_amount

        sale_lines = self._get_sale_lines_with_seat(
            order=order,
            seat=seat,
        )

        for sale_line in sale_lines:
            line_id = get_line_id_from_sale_line(sale_line=sale_line)
            if line_id not in discount_line_values[FAKE_DISCOUNT_ID]:
                continue

            amount += discount_line_values[FAKE_DISCOUNT_ID][line_id]

        return amount

    def _get_total_gross_amount_with_restrictions(
        self,
        order,          # type: Order
        restriction,    # type: Restrictions
        seat=None,      # type: Optional[int]
    ):
        # type: (...) -> float

        if not restriction.exists() and not seat:
            return float(order.total_gross)

        return self._get_total_amount_with_restrictions(
            order=order,
            restriction=restriction,
            seat=seat,
        )

    def _get_order_max_discount_amount(
        self,
        order,          # type: Order
        restriction,    # type: Restrictions
        seat=None,      # type: Optional[int]
    ):
        # type: (...) -> float

        order_xml = order.order_picture
        if order_xml is None:
            return 0

        due_total_amount = float(order_xml.get("dueAmount") or 0)
        if due_total_amount > 0:
            due_total_amount -= 0.01
            due_total_amount = round_value(value=due_total_amount)

        if not restriction.exists() and not seat:
            return due_total_amount

        due_amount = self._get_total_amount_with_restrictions(
            order=order,
            restriction=restriction,
            seat=seat,
        )

        return min(due_amount, due_total_amount)

    def _get_items_to_apply_discount(
            self,
            action,     # type: Action
    ):
        # type: (...) -> Dict

        items_apply_discount = {}
        for entry in action.value:
            key = int(entry.get("item", -1))

            items_apply_discount = self._manager_item_to_apply_discount(items_apply_discount, key, entry)

        return items_apply_discount

    def _get_items_by_groups_to_apply_discount(
        self,
        action,                 # type: Action
        use_percentage=False,   # type: Optional[bool]
    ):
        # type: (...) -> (dict, dict)

        items_groups_to_apply_discount = {}
        discount_by_group = None if use_percentage else {}

        for entry in action.value:
            group = entry.get("group", "")
            if group not in self.product_groups:
                continue

            if not use_percentage:
                quantity = int(entry.get("quantity", 0))
                quantity = min(action.times_to_apply, quantity) if action.times_to_apply else quantity
                discount_by_group[group] = dict(amount=float(entry.get("amount", 0)), quantity=quantity)

            for product_id in self.product_groups[group]:
                key = int(product_id)
                items = self._manager_item_to_apply_discount(items_groups_to_apply_discount, key, entry, group)
                items_groups_to_apply_discount = items

        return items_groups_to_apply_discount, discount_by_group

    @staticmethod
    def _manager_item_to_apply_discount(
            items_to_apply_discount,  # type: Dict
            item_id,                  # type: int
            entry,                    # type: Dict,
            group="",                 # type: Optional[str],
    ):
        # type: (...) -> Dict

        amount = float(entry.get("amount", entry.get("percentage", 0)))
        quantity = int(entry.get("quantity", 0))

        if item_id not in items_to_apply_discount:
            items_to_apply_discount[item_id] = {"quantity": quantity, "amount": amount,  "group": group}

        return items_to_apply_discount

    def _get_amount(
        self,
        amount,             # type: float
        order,              # type: Order
        restriction,        # type: Restrictions
        use_percentage,     # type: bool
        seat,               # type: str
        persistent_value,   # type: float
    ):
        # type: (...) -> float

        if amount == 0:
            raise InvalidAmountException()

        total_gross = self._get_total_gross_amount_with_restrictions(
            order=order,
            restriction=restriction,
            seat=seat,
        )
        if use_percentage:
            amount = total_gross * (amount / 100.0)

        amount = persistent_value if persistent_value is not None else amount
        return min(amount, total_gross)

    @staticmethod
    def _get_quantity_to_apply_discount(
        sale_lines,     # type: List[SaleLine]
        conditions,     # type: Conditions
    ):
        # type: (...) -> int

        necessary_items = {}
        for required_item in conditions.required_items_on_sale:
            necessary_items[str(required_item.item_id)] = required_item.quantity

        order_necessary_items = {}
        for sale_line in sale_lines:
            if sale_line.item_id in necessary_items:
                if sale_line.item_id not in order_necessary_items:
                    order_necessary_items[sale_line.item_id] = 0

                order_necessary_items[sale_line.item_id] += sale_line.qty

        quantity = None
        for key, value in order_necessary_items.iteritems():
            x = value / necessary_items[key]
            quantity = min(quantity, x) if quantity else x

        return quantity

    def _get_quantity_to_apply(self, burnt_items, quantity, uow):
        # type: (Dict, int, AddBenefitToApplyUnitOfWork) -> int

        available_qty = []
        ignore = uow.benefit.conditions.ignore_burned_items
        for required in uow.benefit.conditions.required_items_on_sale:
            available_item_quantity = self._get_available_item_quantity(
                burnt_items={} if ignore else burnt_items,
                required=required,
                uow=uow
            )
            available_qty.append(available_item_quantity / required.quantity)

        for required in uow.benefit.conditions.required_groups_on_sale:
            available_item_quantity = self._get_available_item_quantity(
                burnt_items={} if ignore else burnt_items,
                required=required,
                uow=uow,
                is_group=True
            )
            available_qty.append(available_item_quantity / required.quantity)

        quantity *= min(available_qty) if available_qty else 1

        # TODO: Need to validate rule to handle cases where the item is sold by kg
        if quantity % 1 > 0:
            raise InvalidQuantityException()

        return int(quantity)

    def _get_available_item_quantity(
        self,
        burnt_items,    # type: Dict
        required,       # type: Any[Union[RequiredItemOnSale, str]]
        uow,            # type: AddBenefitToApplyUnitOfWork
        is_group=False  # type: Optional[bool]
    ):
        # type: (...) -> int

        available_quantity = 0
        for sale_line in uow.order.sale_lines:
            valid_context = is_group or required.context is None or sale_line.context == required.context
            item_in_group = is_group and int(sale_line.item_id) in self.product_groups.get(required.group, [])
            required_item_id = "" if is_group else str(required.item_id)
            if valid_context and (item_in_group or sale_line.item_id == required_item_id):
                available_quantity += sale_line.qty
                if sale_line.key in burnt_items:
                    available_quantity -= burnt_items[sale_line.key]

        return available_quantity

    def _do_order_tender(self, pos_id, tender_id, amount):
        # type: (int, int, float) -> Optional[int]

        order_tender = self.pos_ot.doTender(pos_id, tender_id, amount, donotclose=True)
        if order_tender:
            order_tender_id = int(order_tender.get("tenderId", 0) or 0)
            return None if order_tender_id == 0 else order_tender_id
        return None

    @staticmethod
    def _validate_tender(tender_id):
        # type: (int) -> bool

        all_tenders = get_tender_types()
        all_tenders = [key for key in all_tenders.keys()]
        return tender_id in all_tenders

    @staticmethod
    def _get_sale_line_discounts(order):
        sale_line_discounts = {}

        order_picture = order.order_picture
        benefit_list = order_picture.find(".//OrderProperty[@key=\"BENEFIT_LIST\"]")
        benefit_list = benefit_list.get("value") if benefit_list is not None else None

        benefits = json.loads(benefit_list) if benefit_list else {}

        for _, line_detail in benefits.get("discountItemValues", {}).items():
            for line_id, amount in line_detail.items():
                if line_id not in sale_line_discounts.keys():
                    sale_line_discounts[line_id] = amount
                else:
                    sale_line_discounts[line_id] += amount

        return sale_line_discounts

    def _split_discount_by_sale_line(
        self,
        order,                  # type: Order
        restriction,            # type: Restrictions
        discount_id,            # type: str
        discount_amount,        # type: float
        seat=None,              # type: Optional[int]
    ):
        # type: (...) -> Dict

        sale_line_discounts = self._get_sale_line_discounts(order=order)

        discount_by_line_id = {discount_id: {}}
        valid_sale_lines = {}
        total_amount = 0.0
        sale_lines = self._get_sale_lines_with_seat(
            order=order,
            seat=seat,
        )
        for sale_line in sale_lines:
            if restriction.validate(sale_line):
                line_id = get_line_id_from_sale_line(sale_line=sale_line)
                current_discount = sale_line_discounts.get(line_id, 0.0)
                price = (sale_line.unit_price * sale_line.qty) - current_discount

                if price <= 0:
                    continue

                total_amount += price
                valid_sale_lines[line_id] = price

        for key, line_amount in valid_sale_lines.iteritems():
            discount_amount = min(discount_amount, total_amount)
            sale_line_amount = round((discount_amount * (line_amount / total_amount)), 2)
            discount_by_line_id[discount_id][key] = sale_line_amount

        self._distribute_pending_discount_amount(
            order_sale_lines=order.sale_lines,
            sale_line_discounts=sale_line_discounts,
            discount_by_line_id=discount_by_line_id[discount_id],
            discount_amount=discount_amount,
        )

        return discount_by_line_id

    def _distribute_pending_discount_amount(
        self,
        order_sale_lines,               # type: List[SaleLine]
        sale_line_discounts,            # type: Dict[str, float]
        discount_by_line_id,            # type: Dict[str, float]
        discount_amount,                # type: float
    ):
        # type: (...) -> None

        applied_discount = sum(discount_by_line_id.values())
        pending_discount = round_value(value=discount_amount - applied_discount)
        if abs(pending_discount) == 0:
            return

        sale_lines = dict([(get_line_id_from_sale_line(sale_line=s), s) for s in order_sale_lines])
        self._split_and_apply_pending_discounts(
            sale_line_discounts=sale_line_discounts,
            sale_lines=sale_lines,
            discount_pending_to_apply=pending_discount,
            discount_by_line_id=discount_by_line_id,
        )

    @staticmethod
    def _split_and_apply_pending_discounts(
        sale_line_discounts,            # type: Dict[str, float]
        sale_lines,                     # type: Dict[str, SaleLine]
        discount_pending_to_apply,      # type: float
        discount_by_line_id,            # type: Dict[str, float]
    ):
        # type: (...) -> None

        for sale_line_id, _ in discount_by_line_id.items():
            if round_value(value=discount_pending_to_apply) == 0:
                break

            sale_line = sale_lines[sale_line_id]
            real_value = sale_line.item_price - sale_line_discounts.get(sale_line_id, 0)
            remove = discount_pending_to_apply < 0
            discount_limit = discount_by_line_id[sale_line_id] if remove else real_value
            discount_amount = min(discount_limit, abs(discount_pending_to_apply))
            if remove:
                discount_by_line_id[sale_line_id] -= discount_amount
                discount_pending_to_apply += discount_amount
            else:
                discount_by_line_id[sale_line_id] += discount_amount
                discount_pending_to_apply -= discount_amount

    @staticmethod
    def _get_item_price(
        order,                  # type: Order
        sale_line,              # type: SaleLine
        quantity=1,             # type: Optional[int]
        current_price=True,     # type: Optional[str]
    ):
        # type: (...) -> float

        not_combo = sale_line.item_type != CompositionType.COMBO
        if not_combo or sale_line.item_price != 0:
            return (sale_line.current_unit_price if current_price else sale_line.unit_price) * quantity

        combo_price = 0.0
        for line in order.sale_lines:
            if sale_line.line_number != line.line_number:
                continue
            combo_price += (line.current_unit_price if current_price else line.unit_price) * quantity

        return combo_price

    def _filter_sale_lines_by_seat(
        self,
        action,     # type: Action
        order,      # type: Order
    ):
        # type: (...) -> List[SaleLine]

        entry_sale_lines = []
        validate_seat = False
        for entry in action.value:
            entry_seat = entry.get("seat", None)
            if not entry_seat:
                continue
            elif not validate_seat:
                validate_seat = True

            entry_sale_lines.extend(self._get_sale_lines_with_seat(
                order=order,
                seat=entry_seat,
            ))

        if validate_seat:
            return entry_sale_lines
        else:
            return order.sale_lines

    @staticmethod
    def _get_sale_lines_with_seat(
        order,  # type: Order
        seat,   # type: int
    ):
        # type: (...) -> List[SaleLine]

        if not seat:
            return order.sale_lines

        sale_lines = []
        seats_by_line = {}
        for line in order.sale_lines:
            if line.level == '0':
                line_seat = int(line.seat) if line.seat else None
                seats_by_line[line.line_number] = int(line_seat) if line_seat else None
            else:
                line_seat = seats_by_line.get(line.line_number, None)

            if line_seat == seat:
                sale_lines.append(line)

        return sale_lines

    @staticmethod
    def _add_id_if_needs_to_remove_previous_benefits(
        order,                      # type: Order
        line_id,                    # type: str
        benefits_ids_to_remove,     # type: List[str]
        apply_on_total,             # type: bool
    ):
        # type: (...) -> List[str]

        if apply_on_total:
            return benefits_ids_to_remove

        discounts_ids = order.order_discounts.discount_ids.get(line_id, None)
        to_remove = []
        if not discounts_ids:
            return benefits_ids_to_remove

        benefit_list = order.applied_benefits
        for benefit in benefit_list:
            if not benefit.benefit_applied:
                continue

            used_benefits = benefit.used_discount_ids
            if not (set(discounts_ids) & set(used_benefits)):
                continue

            if have_action_by_items_or_groups(benefit=benefit.benefit):
                continue

            if have_action_by_items_or_groups(
                benefit=benefit.benefit,
                on_total=True,
            ):
                item_values_line_ids = get_item_values_line_ids(benefit=benefit)
                if line_id in item_values_line_ids:
                    continue

            to_remove.append(benefit.benefit_key)

        if to_remove:
            benefits_ids_to_remove = list(set(benefits_ids_to_remove + to_remove))

        return benefits_ids_to_remove

    def _get_tender_id(self, action):
        if action.tender_id is None:
            return None

        tender_id = action.tender_id
        if self._validate_tender(tender_id):
            return action.tender_id

        return None



from application.interactor.apply.service import BenefitApplierService
from application.repository.pos.utils import find_benefit_applier
from .dto.request import ApplyCouponRequest


class ApplyCouponInteractor(object):

    def __init__(
        self,
        benefit_applier_service,        # type: BenefitApplierService
    ):
        # type: (...) -> None

        self.benefit_applier_service = benefit_applier_service

    def apply_coupon(
        self,                  # type: ApplyCouponInteractor
        apply_coupon_request,  # type: ApplyCouponRequest
    ):
        # type: (...) -> None

        benefit_applier = find_benefit_applier(
            pos_id=apply_coupon_request.context.originator_pos_id,
            coupon_id=apply_coupon_request.coupon_id
        )

        self.benefit_applier_service.apply_benefit_by_coupon(
            coupon_id=apply_coupon_request.coupon_id,
            context=apply_coupon_request.context,
            benefit_applier=benefit_applier,
        )



from typing import Optional


class Account(object):
    def __init__(
        self,
        account_type,    # type: int
        amount,          # type: Optional[float]
        tender_id,       # type: int
        operation_date,  # type: str
        glaccount,       # type: str
        is_undeclared,   # type: bool
    ):
        # type: (...) -> None

        self.account_type = account_type
        self.amount = amount
        self.tender_id = tender_id
        self.operation_date = operation_date
        self.glaccount = glaccount
        self.is_undeclared = is_undeclared



from domain import OperatorNameRetriever
from old_helper import BaseRepository
from persistence import Connection


class MwOperatorNameRetriever(OperatorNameRetriever, BaseRepository):
    def __init__(self, mbcontext):
        super(MwOperatorNameRetriever, self).__init__(mbcontext)

    def get_operator_name(self, operator_id):
        def inner_func(conn):
            # type: (Connection) -> str
            query = "select LongName from Users where UserId={}".format(operator_id)
            for line in conn.select(query):
                return line.get_entry(0).decode("utf-8")

        return self.execute_with_connection(inner_func)



import logging
from logging import Logger

from cfgtools import read
from helper import config_logger
from model.config import Config, PosMap
from systools import sys_parsefilepath
from typing import List


class Configurator(object):

    def __init__(self, loader_path):
        # type: (str) -> None

        self.config = Config()
        self.config.loader_path = loader_path
        self.reader = read(loader_path)

    def read_loader_configs(self):
        # type: () -> Config

        group = "InternalOrder."
        flask_configs = "FlaskConfiguration."
        totem_configs = "TotemConfigs."

        self.config.component_name = str(self.reader.find_value("Identification.ComponentName"))
        self.config.required_services = "|".join(self.reader.find_values(group + "RequiredServices") or [])
        self.config.host = str(self.reader.find_value(group + flask_configs + "Host"))
        self.config.port = str(self.reader.find_value(group + flask_configs + "Port"))
        self.config.cors_origin_hosts = self.reader.find_values(group + flask_configs + "AllowedOriginHosts")
        self.config.show_coupon_button_on = self.reader.find_values(group + totem_configs + "ShowCouponButtonOn")
        self.config.show_products_scroll = self.reader.find_value(group + totem_configs + "ShowProductsScrollBar")

        if self.config.component_name == "InternalOrderMessageBus":
            return self.config

        self.config.allowed_payment_methods = map(int, self.reader.find_values(group + "AllowedPaymentMethods") or [])
        self.config.available_sale_types = "|".join(self.reader.find_values(group + "AvailableSaleTypes") or [])
        self.config.ruptured_items_cache_time = int(self.reader.find_value(group + "RupturedItemsCacheTime") or 10)
        self.config.api_key_mapping = self.build_pos_map(group)
        self.config.images_path = self._get_file_path(group, "ImagesPath")
        self.config.totem_config_path = self._get_file_path(group, "TotemConfigPath")
        self.config.up_sell_part_codes_path = self._get_file_path(group, "UpSellPartCodes")

        cache_group = "{}RefreshCacheTime.".format(group)
        refresh_cache_time = dict(
            menu=int(self.reader.find_value(cache_group + "Menu") or 5),
            fiscal=int(self.reader.find_value(cache_group + "Fiscal") or 5),
            manager=int(self.reader.find_value(cache_group + "Manager") or 5),
            printer=int(self.reader.find_value(cache_group + "Printer") or 5),
            eft=int(self.reader.find_value(cache_group + "Eft") or 5),
        )

        self.config.refresh_cache_time = refresh_cache_time

        return self.config

    def configure_logger(self):
        # type: () -> Logger

        config_logger(self.config.loader_path, self.config.component_name, root_level=logging.DEBUG)
        return logging.getLogger(self.config.component_name)

    def build_pos_map(self, group):
        # type: (str) -> List[PosMap]

        mapped_pos = []
        to_map_groups = self.reader.find_group(group + "PosMap")

        for group in to_map_groups.groups:
            pos_map = PosMap()
            pos_map.pos_id = int(group.name)
            pos_map.api_key = group.find_value("ApiKey")
            pos_map.abstract_pos_id = int(group.find_value("AbstractPosId"))
            pos_map.user_id = int(group.find_value("UserId"))
            pos_map.user_password = int(group.find_value("UserPassword"))

            mapped_pos.append(pos_map)

        return mapped_pos

    def _get_file_path(self, group, path_key):
        file_path = self.reader.find_value(group + path_key) or ""
        file_path = sys_parsefilepath(file_path) if file_path else file_path

        return file_path



from typing import (
    List,
    Optional,
)
from xml.etree import cElementTree as eTree

from application.interactor import BurnBenefitInteractor, UnlockBenefitInteractor
from application.model import BenefitControllerDto, OperationDescription
from application.model.configuration import Configurations
from application.repository.pos import DBRepository


class BenefitRetryInteractor(object):

    def __init__(self, configs, db_repository, burn_benefit_interactor, unlock_benefit_interactor):
        # type: (Configurations, DBRepository, BurnBenefitInteractor, UnlockBenefitInteractor) -> None

        self.configs = configs
        self.logger = self.configs.logger
        self.db_repository = db_repository
        self.burn_interactor = burn_benefit_interactor
        self.unlock_interactor = unlock_benefit_interactor

    def get_pending_operations_benefits(self):
        # type: () -> Optional[List[BenefitControllerDto]]

        return self.db_repository.find_pending_operations()

    def perform_operations_on_pending_benefits(self, pending_operations):
        # type: (List[BenefitControllerDto]) -> None

        for benefit in pending_operations:
            if not benefit.order_picture:
                self.db_repository.change_benefit_operation_status_to_error(
                    benefit.benefit_id
                )
                continue

            order_picture = eTree.XML(benefit.order_picture)
            pos_id = order_picture.get("posId")

            if benefit.operation_description == OperationDescription.BURNED:
                self.burn_interactor.burn_benefit(
                    benefit_id=benefit.benefit_id,
                    originator_pos_id=pos_id,
                    benefit_applier=benefit.benefit_applier,
                    order_picture=eTree.XML(benefit.order_picture),
                    table_picture=eTree.XML(benefit.table_picture) if benefit.table_picture else None,
                    retry_quantity=benefit.retry_quantity,
                )

            elif benefit.operation_description == OperationDescription.UNLOCKED:
                self.unlock_interactor.unlock_benefit(
                    benefit_id=benefit.benefit_id,
                    originator_pos_id=pos_id,
                    benefit_applier=benefit.benefit_applier,
                    order_picture=eTree.XML(benefit.order_picture),
                    table_picture=eTree.XML(benefit.table_picture) if benefit.table_picture else None,
                    retry_quantity=benefit.retry_quantity,
                )



class ConfirmFinishPaymentData(object):
    def __init__(
        self,
        pos_id,     # type: str
        status      # type: str
    ):
        # type: (...) -> None

        self.pos_id = int(pos_id)
        self.status = int(status)

    def __repr__(self):
        return """ConfirmFinishPaymentData(pos_id={}, status={})""".format(
            self.pos_id,
            self.status
        )

    def __str__(self):
        return self.__repr__()



# encoding: utf-8
from typing import Any


class CardBrand(object):
    def __init__(
        self,
        row_database    # type: Any
    ):
        # type: (...) -> None

        self.brand = row_database.get_entry("Bandeira")
        self.description = row_database.get_entry("Descricao")



from abc import ABCMeta, abstractmethod

from typing import List

from ._Event import Event
from ._Message import Message
from ._SubscriptionInfo import SubscriptionInfo


class MessageBus(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def publish_event(self, event):
        # type: (Event) -> Message
        raise NotImplementedError()

    @abstractmethod
    def send_message(self, component, message):
        # type: (str, Message) -> Message
        raise NotImplementedError()

    @abstractmethod
    def subscribe(self, subscriptions):
        # type: (List[SubscriptionInfo]) -> None
        raise NotImplementedError()

    @abstractmethod
    def reply_message(self, message, reply_message):
        # type: (Message, Message) -> None
        raise NotImplementedError()

    @abstractmethod
    def reply_event(self, event, reply_message):
        # type: (Event, Message) -> None
        raise NotImplementedError()



# ASN.1 named integers
from pyasn1 import error

__all__ = [ 'NamedValues' ]

class NamedValues:
    def __init__(self, *namedValues):
        self.nameToValIdx = {}; self.valToNameIdx = {}
        self.namedValues = ()        
        automaticVal = 1
        for namedValue in namedValues:
            if isinstance(namedValue, tuple):
                name, val = namedValue
            else:
                name = namedValue
                val = automaticVal
            if name in self.nameToValIdx:
                raise error.PyAsn1Error('Duplicate name %s' % (name,))
            self.nameToValIdx[name] = val
            if val in self.valToNameIdx:
                raise error.PyAsn1Error('Duplicate value %s=%s' % (name, val))
            self.valToNameIdx[val] = name
            self.namedValues = self.namedValues + ((name, val),)
            automaticVal = automaticVal + 1

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, ', '.join([repr(x) for x in self.namedValues]))

    def __str__(self): return str(self.namedValues)

    def __eq__(self, other): return tuple(self) == tuple(other)
    def __ne__(self, other): return tuple(self) != tuple(other)
    def __lt__(self, other): return tuple(self) < tuple(other)
    def __le__(self, other): return tuple(self) <= tuple(other)
    def __gt__(self, other): return tuple(self) > tuple(other)
    def __ge__(self, other): return tuple(self) >= tuple(other)
    def __hash__(self): return hash(tuple(self))
    
    def getName(self, value):
        if value in self.valToNameIdx:
            return self.valToNameIdx[value]

    def getValue(self, name):
        if name in self.nameToValIdx:
            return self.nameToValIdx[name]
    
    def __getitem__(self, i): return self.namedValues[i]
    def __len__(self): return len(self.namedValues)

    def __add__(self, namedValues):
        return self.__class__(*self.namedValues + namedValues)
    def __radd__(self, namedValues):
        return self.__class__(*namedValues + tuple(self))
        
    def clone(self, *namedValues):
        return self.__class__(*tuple(self) + namedValues)

# XXX clone/subtype?



from __future__ import absolute_import
from future.utils import PY2, PY26

from subprocess import *

if PY2:
    __future_module__ = True
    from commands import getoutput, getstatusoutput

if PY26:
    from future.backports.misc import check_output



import logging

from sysactions import send_message
from msgbus import TK_SYS_ACK, FM_PARAM
from model.customexception import ErrorGettingNFType, ErrorGettingCertificateExpirationDays
from bustoken import TK_FISCALWRAPPER_GET_NF_TYPE, TK_FISCALWRAPPER_GET_CERTIFICATE_EXPIRATION_DAYS


class FiscalWrapperInteractor(object):

    def __init__(self):
        self.logger = logging.getLogger("PosActions")

    @staticmethod
    def get_nf_type():
        # type: () -> str

        response = send_message("FiscalWrapper", TK_FISCALWRAPPER_GET_NF_TYPE, FM_PARAM, None, timeout=30000000)
        if response.token == TK_SYS_ACK:
            return response.data.upper()

        raise ErrorGettingNFType(response.data)

    @staticmethod
    def get_certificate_expiration_days():
        # type: () -> int

        response = send_message(
            dest_name="FiscalWrapper",
            token=TK_FISCALWRAPPER_GET_CERTIFICATE_EXPIRATION_DAYS,
            format=FM_PARAM,
            data='',
            timeout=30000000,
        )

        if response.token == TK_SYS_ACK:
            return int(response.data)

        raise ErrorGettingCertificateExpirationDays(response.data)



from application.service import BenefitService
from application.model.transactions import AddBenefitToApplyUnitOfWork


class AddBenefitToApplyInteractor:
    def __init__(self, benefit_service):
        # type: (BenefitService) -> None

        self.benefit_service = benefit_service

    def apply(self, uow):
        # type: (AddBenefitToApplyUnitOfWork) -> None

        self.benefit_service.fix_and_apply_benefit(
            uow=uow,
            is_manual=True,
        )
        self.benefit_service.send_order_update_event(static_uow=uow.static_uow)



from datetime import datetime  # noqa

from typing import List  # noqa
from typing import Union  # noqa

from _OrderTender import OrderTender  # noqa


class OrderSate(object):
    VOIDED = 4
    PAID = 5


class Order(object):
    def __init__(self, order_id, date, state, total, donation, tenders, void_reason_id, change_amount):
        # type: (int, datetime, int, float, float, List[OrderTender], Union[int, None], float) -> None
        self.order_id = order_id
        self.date = date
        self.state = state
        self.total = total
        self.donation = donation
        self.tenders = tenders
        self.void_reason_id = void_reason_id
        self.change_amount = change_amount



from typing import Dict


class Exclusions(object):
    def __init__(
        self,
        obj,  # type: Dict
    ):
        # type: (...) -> None

        self.items = list(obj.get("items")) if obj.get("items") else None
        self.groups = list(obj.get("groups")) if obj.get("groups") else None



from ._StaticUnitOfWork import StaticUnitOfWork
from ._AddBenefitToApplyUnitOfWork import AddBenefitToApplyUnitOfWork
from ._VerifyAndRemoveBenefitsUnitOfWork import VerifyAndRemoveBenefitsUnitOfWork



# encoding: utf-8

from typing import Any, Dict

from _OrderBenefitInfo import OrderBenefitInfo


class OrderBenefit(object):
    def __init__(self, row_database, discount_items_values, list_benefit):
        # type: (Any, Dict, OrderBenefitInfo) -> OrderBenefit

        self.order_id = row_database.get_entry("OrderId")
        self.business_period = row_database.get_entry("BusinessPeriod")
        self.part_code = row_database.get_entry("PartCode")
        self.item_id = row_database.get_entry("ItemId")
        self.level = row_database.get_entry("Level")
        self.product_name = row_database.get_entry("ProductName")
        self.ordered_qty = float(row_database.get_entry("OrderedQty"))
        self.line_number = row_database.get_entry("LineNumber")
        self.default_unit_price = float(row_database.get_entry("DefaultUnitPrice"))
        self.discount_value = 0.0
        self.discount_type = row_database.get_entry("discountType")
        self.key_burn = "{}.{}.{}.{}".format(self.line_number, self.level, self.item_id, self.part_code)
        self.discount_items_values = discount_items_values
        self.applied_benefits = list_benefit



from pos_model import TenderType


def is_electronic_tender_id(
    tender_id,      # type: int
):
    electronic_tender_ids = [
        TenderType.credit,
        TenderType.debit,
        TenderType.voucher,
        TenderType.digital_wallet,
        TenderType.gift_card,
    ]
    return int(tender_id) in electronic_tender_ids



from messagebus import TokenCreator, TokenPriority


class BusTokens(object):
    TK_AUDITLOG_RESYNC = TokenCreator.create_token(TokenPriority.high, "14", "1")
    TK_SALE_ORDER_MODIFIED = TokenCreator.create_token(TokenPriority.low, "A", "32")



"""
This module contains backports the data types that were significantly changed
in the transition from Python 2 to Python 3.

- an implementation of Python 3's bytes object (pure Python subclass of
  Python 2's builtin 8-bit str type)
- an implementation of Python 3's str object (pure Python subclass of
  Python 2's builtin unicode type)
- a backport of the range iterator from Py3 with slicing support

It is used as follows::

    from __future__ import division, absolute_import, print_function
    from builtins import bytes, dict, int, range, str

to bring in the new semantics for these functions from Python 3. And
then, for example::
    
    b = bytes(b'ABCD')
    assert list(b) == [65, 66, 67, 68]
    assert repr(b) == "b'ABCD'"
    assert [65, 66] in b

    # These raise TypeErrors:
    # b + u'EFGH'
    # b.split(u'B')
    # bytes(b',').join([u'Fred', u'Bill'])


    s = str(u'ABCD')

    # These raise TypeErrors:
    # s.join([b'Fred', b'Bill'])
    # s.startswith(b'A')
    # b'B' in s
    # s.find(b'A')
    # s.replace(u'A', b'a')

    # This raises an AttributeError:
    # s.decode('utf-8')

    assert repr(s) == 'ABCD'      # consistent repr with Py3 (no u prefix)


    for i in range(10**11)[:10]:
        pass

and::
    
    class VerboseList(list):
        def append(self, item):
            print('Adding an item')
            super().append(item)        # new simpler super() function

For more information:
---------------------

- future.types.newbytes
- future.types.newdict
- future.types.newint
- future.types.newobject
- future.types.newrange
- future.types.newstr


Notes
=====

range()
-------
``range`` is a custom class that backports the slicing behaviour from
Python 3 (based on the ``xrange`` module by Dan Crosta). See the
``newrange`` module docstring for more details.


super()
-------
``super()`` is based on Ryan Kelly's ``magicsuper`` module. See the
``newsuper`` module docstring for more details.


round()
-------
Python 3 modifies the behaviour of ``round()`` to use "Banker's Rounding".
See http://stackoverflow.com/a/10825998. See the ``newround`` module
docstring for more details.

"""

from __future__ import absolute_import, division, print_function

import functools
from numbers import Integral

from future import utils


# Some utility functions to enforce strict type-separation of unicode str and
# bytes:
def disallow_types(argnums, disallowed_types):
    """
    A decorator that raises a TypeError if any of the given numbered
    arguments is of the corresponding given type (e.g. bytes or unicode
    string).

    For example:

        @disallow_types([0, 1], [unicode, bytes])
        def f(a, b):
            pass

    raises a TypeError when f is called if a unicode object is passed as
    `a` or a bytes object is passed as `b`.

    This also skips over keyword arguments, so 

        @disallow_types([0, 1], [unicode, bytes])
        def g(a, b=None):
            pass

    doesn't raise an exception if g is called with only one argument a,
    e.g.:

        g(b'Byte string')

    Example use:

    >>> class newbytes(object):
    ...     @disallow_types([1], [unicode])
    ...     def __add__(self, other):
    ...          pass

    >>> newbytes('1234') + u'1234'      #doctest: +IGNORE_EXCEPTION_DETAIL 
    Traceback (most recent call last):
      ...
    TypeError: can't concat 'bytes' to (unicode) str
    """

    def decorator(function):

        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            # These imports are just for this decorator, and are defined here
            # to prevent circular imports:
            from .newbytes import newbytes
            from .newint import newint
            from .newstr import newstr

            errmsg = "argument can't be {0}"
            for (argnum, mytype) in zip(argnums, disallowed_types):
                # Handle the case where the type is passed as a string like 'newbytes'.
                if isinstance(mytype, str) or isinstance(mytype, bytes):
                    mytype = locals()[mytype]

                # Only restrict kw args only if they are passed:
                if len(args) <= argnum:
                    break

                # Here we use type() rather than isinstance() because
                # __instancecheck__ is being overridden. E.g.
                # isinstance(b'abc', newbytes) is True on Py2.
                if type(args[argnum]) == mytype:
                    raise TypeError(errmsg.format(mytype))

            return function(*args, **kwargs)
        return wrapper
    return decorator


def no(mytype, argnums=(1,)):
    """
    A shortcut for the disallow_types decorator that disallows only one type
    (in any position in argnums).

    Example use:

    >>> class newstr(object):
    ...     @no('bytes')
    ...     def __add__(self, other):
    ...          pass

    >>> newstr(u'1234') + b'1234'     #doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
      ...
    TypeError: argument can't be bytes

    The object can also be passed directly, but passing the string helps
    to prevent circular import problems.
    """
    if isinstance(argnums, Integral):
        argnums = (argnums,)
    disallowed_types = [mytype] * len(argnums)
    return disallow_types(argnums, disallowed_types)


def issubset(list1, list2):
    """
    Examples:

    >>> issubset([], [65, 66, 67])
    True
    >>> issubset([65], [65, 66, 67])
    True
    >>> issubset([65, 66], [65, 66, 67])
    True
    >>> issubset([65, 67], [65, 66, 67])
    False
    """
    n = len(list1)
    for startpos in range(len(list2) - n + 1):
        if list2[startpos:startpos+n] == list1:
            return True
    return False


if utils.PY3:
    import builtins
    bytes = builtins.bytes
    dict = builtins.dict
    int = builtins.int
    list = builtins.list
    object = builtins.object
    range = builtins.range
    str = builtins.str

    # The identity mapping
    newtypes = {bytes: bytes,
                dict: dict,
                int: int,
                list: list,
                object: object,
                range: range,
                str: str}

    __all__ = ['newtypes']

else:

    from .newbytes import newbytes
    from .newdict import newdict
    from .newint import newint
    from .newlist import newlist
    from .newrange import newrange
    from .newobject import newobject
    from .newstr import newstr

    newtypes = {bytes: newbytes,
                dict: newdict,
                int: newint,
                long: newint,
                list: newlist,
                object: newobject,
                range: newrange,
                str: newbytes,
                unicode: newstr}

    __all__ = ['newbytes', 'newdict', 'newint', 'newlist', 'newrange', 'newstr', 'newtypes']




from _OpenDaySummaryHeaderDto import OpenDaySummaryHeaderDto  # noqa
from _OpenDaySummaryBodyDto import OpenDaySummaryBodyDto  # noqa


class OpenDaySummaryDto(object):
    def __init__(self, open_day_summary_header_dto, open_day_summary_body_dto):
        # type:(OpenDaySummaryHeaderDto, OpenDaySummaryBodyDto) -> ()
        self.open_day_summary_header_dto = open_day_summary_header_dto
        self.open_day_summary_body_dto = open_day_summary_body_dto



# -*- coding: utf-8 -*-
"""
python-escpos enables you to manipulate escpos-printers
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__all__ = ["constants", "escpos", "exceptions", "printer"]

try:
    from .version import version as __version__  # noqa
except ImportError:  # pragma: no cover
    raise ImportError(
        'Failed to find (autogenerated) version.py. '
        'This might be because you are installing from GitHub\'s tarballs, '
        'use the PyPI ones.'
    )



class FiscalStatusDto(object):
    def __init__(
            self,
            status,     # type: bool
            message,    # type: str
    ):
        # type: (...) -> None

        self.status = status
        self.message = message



import datetime

from commons.dto import ReportColumnDefinition, AlignTypes
from commons.report import Formatter
from commons.util import TableReport
from reports_app.cashtransfersummary.dto import CashTransferSummaryDto  # noqa
from reports_app.cashtransfersummary.dto import CashTransferSummaryHeaderDto  # noqa


class CashTransferFormatter(Formatter):
    def __init__(self):
        self.report_lines = []
        self.translated_messages = []

    def format_report(self, cash_transfer_summary_dto, translator):
        # type: (CashTransferSummaryDto) -> None

        self.translated_messages = translator.translate_labels(
            labels=[
                "$CASH_OUT_TITLE",
                "$CASH_IN_TITLE",
                "$BUSINESS_DATE",
                "$STORE",
                "$LABEL_POS",
                "$OPERATOR",
                "$AUTHORIZER",
                "$CURRENT_DATE",
                "$BUSINESS_DATE",
                "$ENVELOPE_NUMBER",
                "$VALUE",
            ],
            pos_id=cash_transfer_summary_dto.cash_transfer_summary_header_dto.pos_id,
        )

        report = self._append_header(cash_transfer_summary_dto.cash_transfer_summary_header_dto)
        report += self._append_body(cash_transfer_summary_dto.cash_transfer_summary_body_dto)

        report_bytes = report.decode("utf-8").encode("utf-8")
        return report_bytes

    def _append_header(self, cash_transfer_summary_header_dto):
        # type: (CashTransferSummaryHeaderDto) -> str
        self._append_line_separator()
        self._append_title(cash_transfer_summary_header_dto.cash_out)
        self._append_store_pos(cash_transfer_summary_header_dto.store_id, cash_transfer_summary_header_dto.pos_id)
        self._append_operator(cash_transfer_summary_header_dto.operator_id)
        self._append_current_date()
        self._append_business_date(cash_transfer_summary_header_dto.business_date)

        table_formatter = TableReport([
            ReportColumnDefinition(width=17, fill_with_char='.'),
            ReportColumnDefinition(width=21, before_text=': ')
        ])

        return table_formatter.format(self.report_lines)

    def _append_body(self, cash_transfer_summary_body_dto):
        self.report_lines = []

        self._append_empty_line()

        if cash_transfer_summary_body_dto.cash_out:
            self._append_authorizer(cash_transfer_summary_body_dto.authorizer)
            self._append_envelope_number(cash_transfer_summary_body_dto.envelope_number)

        self._append_value(cash_transfer_summary_body_dto.value)
        self._append_line_separator()

        table_formatter = TableReport([
            ReportColumnDefinition(width=17, fill_with_char='.'),
            ReportColumnDefinition(width=21, before_text=': ')
        ])

        return table_formatter.format(self.report_lines)

    def _append_line_separator(self):
        self.report_lines.append([ReportColumnDefinition(
            fill_with_char='='
        )])

    def _append_title(self, cash_in_cash_out):
        title = self.translated_messages["$CASH_OUT_TITLE"] if \
            (cash_in_cash_out == 'cash_out' or cash_in_cash_out is True) else self.translated_messages["$CASH_IN_TITLE"]

        self.report_lines.append([ReportColumnDefinition(
            text=title,
            align=AlignTypes.CENTER
        )])

    def _append_store_pos(self, store_id, pos_id):
        self.report_lines.append([self.translated_messages["$STORE"] + "/" + self.translated_messages["$LABEL_POS"],
                                  (str(store_id) + '/' + str(pos_id))])

    def _append_operator(self, operator_id):
        self.report_lines.append([self.translated_messages["$OPERATOR"], str(operator_id)])

    def _append_current_date(self):
        self.report_lines.append([self.translated_messages["$CURRENT_DATE"],
                                  datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S")])

    def _append_business_date(self, business_date):
        self.report_lines.append([self.translated_messages["$BUSINESS_DATE"], business_date])

    def _append_empty_line(self):
        self.report_lines.append([ReportColumnDefinition(fill_with_char=' ')])

    def _append_authorizer(self, authorizer):
        self.report_lines.append([self.translated_messages["$AUTHORIZER"], str(authorizer)])

    def _append_envelope_number(self, envelope_number):
        self.report_lines.append([self.translated_messages["$ENVELOPE_NUMBER"], str(envelope_number)])

    def _append_value(self, value):
        self.report_lines.append([self.translated_messages["$VALUE"], ("R$ " + str(value))])



# Copyright 2009-2017 Wander Lairson Costa
# Copyright 2009-2020 PyUSB contributors
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# All the hacks necessary to assure compatibility across all
# supported versions come here.
# Please, note that there is one version check for each
# hack we need to do, this makes maintenance easier... ^^

import sys
import array

__all__ = ['_reduce', '_set', '_next', '_update_wrapper']

# we support Python >= 2.4
assert sys.hexversion >= 0x020400f0

# On Python 3, reduce became a functools module function
try:
    import functools
    _reduce = functools.reduce
except (ImportError, AttributeError):
    _reduce = reduce

# all, introduced in Python 2.5
try:
    _all = all
except NameError:
    _all = lambda iter_ : _reduce( lambda x, y: x and y, iter_, True )

# we only have the builtin set type since 2.5 version
try:
    _set = set
except NameError:
    import sets
    _set = sets.Set

# On Python >= 2.6, we have the builtin next() function
# On Python 2.5 and before, we have to call the iterator method next()
def _next(iter):
    try:
        return next(iter)
    except NameError:
        return iter.next()

# functools appeared in 2.5
try:
    import functools
    _update_wrapper = functools.update_wrapper
except (ImportError, AttributeError):
    def _update_wrapper(wrapper, wrapped):
        wrapper.__name__ = wrapped.__name__
        wrapper.__module__ = wrapped.__module__
        wrapper.__doc__ = wrapped.__doc__
        wrapper.__dict__ = wrapped.__dict__

# this is used (as of May 2015) twice in core, once in backend/openusb, and in
# some unit test code. It would probably be clearer if written in terms of some
# definite 3.2+ API (bytearrays?) with a fallback provided for 2.4+.
def as_array(data=None):
    if data is None:
        return array.array('B')

    if isinstance(data, array.array):
        return data

    try:
        return array.array('B', data)
    except TypeError:
        # When you pass a unicode string or a character sequence,
        # you get a TypeError if the first parameter does not match
        a = array.array('B')
        a.fromstring(data) # deprecated since 3.2
        return a



import json

from custom_sysactions import user_authorized_to_execute_action
from sysactions import action

from .service_period_commons import get_selection_service_period


@action
@user_authorized_to_execute_action
def do_service_period_table(pos_id):
    # type: (int) -> bool

    from msgbusboundary import MsgBusTableService as tableService
    from table_actions import create_table_modified_event

    selected_table = tableService().get_selected_table(pos_id)
    if not selected_table:
        return False

    service_period = get_selection_service_period(pos_id)
    if service_period is None:
        return False

    service_period_json = json.dumps(service_period)
    tableService().set_table_custom_property(pos_id, "SERVICE_PERIOD_TABLE", service_period_json, selected_table.id)
    table_picture = tableService().get_table_picture(pos_id, selected_table.id)

    create_table_modified_event(pos_id, table_picture)
    return True



from logging import Logger

from model.error_type import PosValidationErrorsType
from interactor.route.order.get_detail.dto import OrderGetDetailRequest
from model.config import Config
from model.exception import ValidationError
from model.order_detail import OrderDetail
from model.repository import PosRepositoryModel
from model.services import OrderServiceModel
from validation_util import validate_api_key


class OrderGetDetailInteractor(object):
    def __init__(self, logger, config, pos_repository, order_service):
        # type: (Logger, Config, PosRepositoryModel, OrderServiceModel) -> None

        self.logger = logger
        self.config = config
        self.pos_repository = pos_repository
        self.order_service = order_service

    def execute(self, request):
        # type: (OrderGetDetailRequest) -> OrderDetail

        pos_id = validate_api_key(request, self.config.api_key_mapping)
        model = self.pos_repository.get_model(pos_id)
        pos_ot = self.pos_repository.get_pos_ot(model)

        order = self.pos_repository.get_current_order(model)

        if not order and not request.order_id:
            raise ValidationError(PosValidationErrorsType.OrderNotFound)

        order_id = request.order_id if request.order_id else order.get('orderId')

        if not order_id:
            raise ValidationError(PosValidationErrorsType.ParamsNotFound)

        order = self.pos_repository.get_order_details(pos_id, order_id, pos_ot)
        if order:
            return self.order_service.convert_mw_order_to_order_detail(order)

        raise ValidationError(PosValidationErrorsType.InvalidOrderId)



# -*- coding: utf-8 -*-
import json
import locale
import logging
from datetime import datetime
from xml.etree import cElementTree as eTree

import sysactions
from pos_model import TenderType
from repository import OrderRepository, AccountRepository, TenderRepository, FiscalRepository, PosCtrlRepository, \
    ProductRepository, TableRepository
from sysactions import get_storewide_config
from typing import Dict, List, Optional, Tuple, Union

from utils import is_electronic_tender_id
from .model import Order, TotalTender, Transfer
from utils.report_formatters import translate_pos_label

DOUBLE_LINE_SEPARATOR = u"======================================\n"
PAYMENT_TEXT = u"  {0:.<15}: [{1:>4}] R${2:>10}\n"

logger = logging.getLogger("ReportActions")


class CashReport(object):
    TYPE_REAL_DATE = "RealDate"
    TYPE_BUSINESS_PERIOD = "BusinessPeriod"
    TYPE_SESSION_ID = "SessionId"
    TYPE_XML = "Xml"

    def __init__(self, order_repository, account_repository, tender_repository, fiscal_repository, pos_ctrl_repository, table_repository, product_repository, store_id):
        # type: (OrderRepository, AccountRepository, TenderRepository, FiscalRepository, PosCtrlRepository, TableRepository, ProductRepository, unicode, Union[List[str], None]) -> None

        self.order_repository = order_repository
        self.account_repository = account_repository
        self.tender_repository = tender_repository
        self.fiscal_repository = fiscal_repository
        self.pos_ctrl_repository = pos_ctrl_repository
        self.table_repository = table_repository
        self.product_repository = product_repository
        self.store_id = store_id
        self.print_card_brands = get_storewide_config("Store.PrintCardBrands", defval="False").lower() == "true"

        self.tender_names_dict = {}  # type: Dict[int, unicode]
        for tender_tuple in self.tender_repository.get_tender_names():
            self.tender_names_dict[tender_tuple[0]] = tender_tuple[1]

        try:
            locale.setlocale(locale.LC_ALL, "portuguese_brazil")
        except Exception:
            locale.setlocale(locale.LC_ALL, "pt_BR.UTF-8")

    def generate_cash_report_by_real_date(self, pos_id, report_pos, initial_date, end_date, operator_id, voided_lines, business_period=None, codigo_centro=None, close_time=None, is_table_service=False, only_body=False, session_id=None, only_payments=False, is_integrated_delivery=False):
        if business_period is None:
            return self._generate_cash_report(self.TYPE_REAL_DATE, pos_id, report_pos, initial_date, end_date, operator_id, session_id, voided_lines, is_table_service=is_table_service, only_body=only_body, only_payments=only_payments, is_integrated_delivery=is_integrated_delivery)
        else:
            return self._generate_cash_report_json(pos_id, report_pos, initial_date, end_date, operator_id, codigo_centro, close_time, voided_lines)

    def generate_cash_report_by_business_period(self, pos_id, report_pos, initial_date, end_date, operator_id, voided_lines, is_table_service=False, only_body=False, only_payments=False):
        return self._generate_cash_report(self.TYPE_BUSINESS_PERIOD, pos_id, report_pos, initial_date, end_date, operator_id, None, voided_lines, is_table_service=is_table_service, only_body=only_body, only_payments=only_payments)

    def generate_cash_report_by_session_id(self, pos_id, session_id, voided_lines, is_table_service=False, only_body=False, only_payments=False):
        return self._generate_cash_report(self.TYPE_SESSION_ID, pos_id, None, None, None, None, session_id, voided_lines, is_table_service=is_table_service, only_body=only_body, only_payments=only_payments)

    def generate_cash_report_by_xml(self, pos_id, initial_date, end_date, is_table_service=False, only_body=False, only_payments=False):
        return self._generate_cash_report(self.TYPE_XML, pos_id, None, initial_date, end_date, None, None, None, is_table_service=is_table_service, only_body=only_body, only_payments=only_payments)

    def _generate_cash_report(self, report_type, pos_id, report_pos, initial_date, end_date, operator_id, session_id, voided_lines, is_table_service=False, only_body=False, only_payments=False, is_integrated_delivery=False):
        if report_type == self.TYPE_REAL_DATE:
            paid_orders = self.order_repository.get_paid_orders_by_real_date(initial_date, end_date, operator_id, report_pos, session_id)
            voided_orders = self.order_repository.get_voided_orders_by_real_date(initial_date, end_date, operator_id, report_pos, session_id)
            transfers = self.account_repository.get_transfers_by_real_date(initial_date, end_date, operator_id, report_pos, session_id)
            opened_orders = None
            if is_table_service:
                opened_orders = self.order_repository.get_opened_orders(report_type,
                                                                        initial_date=initial_date,
                                                                        end_date=end_date,
                                                                        operator_id=operator_id)

        elif report_type == self.TYPE_BUSINESS_PERIOD:
            paid_orders = self.order_repository.get_paid_orders_by_business_period(initial_date, end_date, operator_id, report_pos)
            voided_orders = self.order_repository.get_voided_orders_by_business_period(initial_date, end_date, operator_id, report_pos)
            transfers = self.account_repository.get_transfers_by_business_period(initial_date, end_date, operator_id, report_pos)
            opened_orders = None
            if is_table_service:
                opened_orders = self.order_repository.get_opened_orders(report_type,
                                                                        initial_date=initial_date,
                                                                        end_date=end_date,
                                                                        operator_id=operator_id)

        elif report_type == self.TYPE_SESSION_ID:
            paid_orders = self.order_repository.get_paid_orders_by_session_id(session_id)
            voided_orders = self.order_repository.get_voided_orders_by_session_id(session_id)
            transfers = self.account_repository.get_transfers_by_session_id(session_id)
            start_index = session_id.find("user=")
            end_index = session_id.find(",", start_index)
            operator_id = session_id[start_index + 5:end_index]
            start_index = session_id.find("period=")
            initial_date = end_date = datetime.strptime(session_id[start_index + 7:], "%Y%m%d")
            opened_orders = None
            if is_table_service:
                opened_orders = self.order_repository.get_opened_orders(report_type,
                                                                        session_id=session_id,
                                                                        operator_id=operator_id)

        else:
            paid_orders = self.order_repository.get_paid_orders_by_xml(initial_date, end_date)
            paid_orders = self._manager_orders_id(paid_orders)
            voided_orders = self.order_repository.get_voided_orders_by_xml(initial_date, end_date)
            transfers = self.account_repository.get_transfers_by_real_date(initial_date, end_date, None, report_pos)
            operator_id = None
            opened_orders = None

        card_brands_payments = self.order_repository.get_payments_by_order(paid_orders)
        payments_by_brand = self.product_repository.populate_card_brands_groups(card_brands_payments)
        tef_card_brands = self._get_tef_card_brands(paid_orders)

        count_pos_error = filter(lambda x: x == "error", paid_orders)
        count_pos_error = len(count_pos_error)

        paid_orders = filter(lambda x: x != "error", paid_orders)
        voided_orders = filter(lambda x: x != "error", voided_orders)

        tables_by_sector = self.table_repository.get_tables_by_sector(paid_orders)
        paid_orders = self.fill_orders_sector(tables_by_sector, paid_orders)
        voided_orders = self.fill_orders_sector(tables_by_sector, voided_orders)

        header = ""
        if not only_body:
            header = self._build_header(report_type, pos_id, report_pos, initial_date, end_date, operator_id, only_payments, count_pos_error, is_integrated_delivery)
        if only_payments:
            body = self._build_body_only_with_payments(payments_by_brand, tef_card_brands)
        else:
            grouped_paid_orders_by_sector = self.get_grouped_paid_orders_by_sector(paid_orders)
            body, additional_info = self._build_body(pos_id, paid_orders, voided_orders, opened_orders, transfers,
                                                     report_type, voided_lines, payments_by_brand, tef_card_brands,
                                                     is_table_service, operator_id, grouped_paid_orders_by_sector)
            if is_table_service and report_type not in [self.TYPE_XML, self.TYPE_SESSION_ID]:
                average_info = self.table_repository.get_sales_average_info(initial_date, end_date, operator_id)
                table_info = self.build_table_info(average_info, grouped_paid_orders_by_sector, additional_info)
                body += table_info

        report = header + body

        report_bytes = report.encode("utf-8")

        return report_bytes

    def _get_tef_card_brands(self, paid_orders):
        grouped_card_brands = {}
        brands_descriptions = self.fiscal_repository.get_brands_description()
        tenders = self.order_repository.get_tender_details(paid_orders)

        for tender_id, tender_amount, tender_detail, order_id in tenders:
            if not is_electronic_tender_id(tender_id) or float(tender_amount) == 0:
                continue

            if not tender_detail:
                message = "Tender detail not found to order_id: {0} and tender_id: {1}". format( order_id, tender_id)
                logger.warning(message)
                continue

            detail = json.loads(tender_detail)
            card_brand = self._get_card_brand(brands_descriptions, detail)
            tender_amount = float(tender_amount)
            tender_id = int(tender_id)
            brand_key = (card_brand, tender_id)
            if brand_key not in grouped_card_brands:
                grouped_card_brands[brand_key] = {}
                grouped_card_brands[brand_key]["tenderId"] = tender_id
                grouped_card_brands[brand_key]["value"] = tender_amount
                grouped_card_brands[brand_key]["quantity"] = 1
            else:
                grouped_card_brands[brand_key]["value"] += tender_amount
                grouped_card_brands[brand_key]["quantity"] += 1

        return grouped_card_brands

    @staticmethod
    def _get_card_brand(brands_descriptions, detail):
        card_brand = detail.get('Bandeira', '-1')
        try:
            card_brand_id = int(card_brand)
        except ValueError:
            return card_brand

        return brands_descriptions[card_brand_id] if card_brand_id in brands_descriptions else 'OUTROS'

    def _build_header(self, report_type, pos_id, report_pos, initial_date, end_date, operator_id, only_payments, count_pos_error=0, is_integrated_delivery=False):
        # type: (unicode, int, Optional[unicode], datetime, datetime, unicode, bool, int, bool) -> unicode
        pos_list = "Todos" if report_pos is None else report_pos.zfill(2)
        pos = translate_pos_label(pos_id=pos_id)
        if report_type == self.TYPE_REAL_DATE:
            report_type_text = "Data Fiscal"
            if is_integrated_delivery:
                report_type_text = "Integradas Delivery"
        elif report_type == self.TYPE_BUSINESS_PERIOD:
            report_type_text = "Dia Negocio"
        elif report_type == self.TYPE_XML:
            report_type_text = "Arquivo Xml"
        else:
            report_type_text = "Surpresa"
            pos_list = pos_id

        if operator_id != "0" and operator_id is not None:
            operator_name = eTree.XML(sysactions.get_user_information(operator_id)).find(".//user").get("LongName")[:20]
        else:
            operator_name = "Todos"

        if not only_payments:
            report_name = "Relatorio Vendas"
        else:
            report_name = "Relatorio Bandeiras"

        header = DOUBLE_LINE_SEPARATOR
        header += u"{0} - {1}".format(report_name, report_type_text).center(38)
        header += u"\nLoja.........: " + self.store_id.upper()
        header += u"\nData/hora....: " + datetime.now().strftime("%d/%m/%Y %H:%M:%S") + "\n"
        if report_type != self.TYPE_SESSION_ID:
            header += u"{0:.<13}: {1} - {2}\n".format("Periodo", initial_date.strftime("%d/%m/%y"), end_date.strftime("%d/%m/%y"))
        header += u"Operador.....: " + u"{0}\n".format(operator_name[:20])
        temp_pos_list = u"{0} incluidos: ".format(pos) + u"{0}".format(pos_list)
        header += temp_pos_list[:38]
        if len(temp_pos_list) > 38:
            header += (u"\n                %s" % temp_pos_list[38:60])
        if len(temp_pos_list) > 66:
            header += (u"\n                %s" % temp_pos_list[60:82])
        header += u"\n"
        if count_pos_error > 0:
            header += u" {0} excluidos: ".format(pos) + u"{0}\n".format(count_pos_error)

        header += DOUBLE_LINE_SEPARATOR

        return header

    def _build_body(self, pos_id, paid_orders, voided_orders, opened_orders, transfers, report_type, voided_lines,
                    payments_by_brand=None, tef_card_brands=None, is_table_service=False, operator_id=None,
                    grouped_paid_orders_by_sector=None):
        # type: (int, List[Order], List[Order], List, List[Transfer], unicode, int, Optional[List[Tuple[unicode, int, float, int]]], dict, bool, int, dict) -> (unicode, list[dict])

        qty_paid_orders, \
        qty_discount, \
        qty_donation, \
        qty_tip, \
        value_paid_orders, \
        value_discount, \
        value_donation, \
        value_tip = self.get_paid_order_values(paid_orders)

        qty_coupons_voided, \
        qty_lines_voided, \
        qty_orders_voided, \
        qty_total_voided, \
        value_coupons_voided, \
        value_lines_voided, \
        value_orders_voided, \
        value_total_voided = self._get_voided_order_values(voided_lines, voided_orders)

        delivery_paid_orders = filter(lambda x: x.sale_type == 3 and x.delivery_originator, paid_orders)
        qty_delivery_orders, value_delivery_orders = self.get_delivery_orders_values(delivery_paid_orders)

        manual_delivery_paid_orders = filter(lambda x: x.sale_type == 3 and not x.delivery_originator, paid_orders)
        qty_manual_delivery_orders, value_manual_delivery_orders = self.get_delivery_orders_values(manual_delivery_paid_orders)

        total_quantity_delivery = qty_delivery_orders + qty_manual_delivery_orders
        total_value_delivery = value_delivery_orders + value_manual_delivery_orders

        total_quantity_table = 0
        total_value_table = 0.0
        lines = []
        for sector in sorted(grouped_paid_orders_by_sector, key=lambda x: grouped_paid_orders_by_sector[x]['value'],
                             reverse=True):
            quantity = grouped_paid_orders_by_sector[sector]['quantity']
            value = self._str_float(grouped_paid_orders_by_sector[sector]['value'])
            total_quantity_table += quantity
            total_value_table += grouped_paid_orders_by_sector[sector]['value']
            lines.append(u"    {0:.<13}: [{1:>4}] R${2:>10}\n".format(sector, quantity, value))

        qty_opened_orders = 0
        value_opened_orders = 0
        if is_table_service:
            qty_opened_orders, value_opened_orders = self.get_opened_orders_values(pos_id, opened_orders)

        initial_balance = self._get_initial_balance(is_table_service, transfers)

        body = u"Descricao          Qtd    Total\n"
        body += DOUBLE_LINE_SEPARATOR
        if not is_table_service:
            body += u"Balanco Inicial..:        R${0:>10}\n".format(self._str_float(initial_balance))
        total_gross = self._str_float(value_paid_orders + value_discount + value_opened_orders + value_tip)
        body += u"Venda Bruta......: [{0:>4}] R${1:>10}\n".format(qty_paid_orders + qty_opened_orders, total_gross)
        body += u"Descontos........: [{0:>4}] R${1:>10}\n".format(qty_discount, self._str_float(value_discount))
        total_qty = self._str_float(value_paid_orders + value_opened_orders + value_tip)
        body += u"Venda Acumulada..: [{0:>4}] R${1:>10}\n".format(qty_paid_orders + qty_opened_orders, total_qty)

        if not (qty_opened_orders != 0 and value_opened_orders == 0) and is_table_service:
            body += u"Venda Aberta.....: [{0:>4}] R${1:>10}\n".format(qty_opened_orders, self._str_float(value_opened_orders))
        total_bill_paid =  value_paid_orders + value_tip
        body += u"Venda Faturada...: [{0:>4}] R${1:>10}\n".format(qty_paid_orders, self._str_float(total_bill_paid))

        total_quantity_counter = qty_paid_orders - total_quantity_table - total_quantity_delivery
        total_value_counter = total_bill_paid - total_value_table - total_value_delivery

        additional_info = [
            {
                "name": "Balcao",
                "value": total_value_counter / total_quantity_counter if total_quantity_counter > 0 else 0,
                "quantity": total_quantity_counter
            },
            {
                "name": "Delivery",
                "value": total_value_delivery / total_quantity_delivery if total_quantity_delivery > 0 else 0,
                "quantity": total_quantity_delivery
            }
        ]

        str_total_value_counter = self._str_float(total_value_counter)
        body += u"  Balcao.........: [{0:>4}] R${1:>10}\n".format(total_quantity_counter, str_total_value_counter)

        str_total_value_table = self._str_float(total_value_table)
        body += u"  Restaurante....: [{0:>4}] R${1:>10}\n".format(total_quantity_table, str_total_value_table)
        for line in lines:
            body += line

        if qty_delivery_orders > 0 or manual_delivery_paid_orders:
            body += u"  Delivery.......: [{0:>4}] R${1:>10}\n".format(total_quantity_delivery,
                                                                      self._str_float(total_value_delivery))

            if manual_delivery_paid_orders:
                body += u"    Manual.......: [{0:>4}] R${1:>10}\n".format(qty_manual_delivery_orders,
                                                                          self._str_float(value_manual_delivery_orders))

            order_originators = sorted(list(dict.fromkeys([x.delivery_originator for x in delivery_paid_orders])))
            for originator in order_originators:
                qty_delivery_brand_orders = len([x for x in delivery_paid_orders
                                                 if x.delivery_originator == originator])
                value_delivery_brand_orders = sum([x.total for x in delivery_paid_orders
                                                   if x.delivery_originator == originator])
                body += u"    {0:.<13}: [{1:>4}] R${2:>10}\n".format(originator[:13].upper(),
                                                                     qty_delivery_brand_orders,
                                                                     self._str_float(value_delivery_brand_orders))

        body += DOUBLE_LINE_SEPARATOR

        print_donation_or_tip = False
        if report_type != self.TYPE_XML and value_donation:
            body += u"Doacoes..........: [{0:>4}] R${1:>10}\n".format(qty_donation, self._str_float(value_donation))
            print_donation_or_tip = True

        if report_type != self.TYPE_XML and qty_tip > 0:
            if operator_id == "0" or operator_id is None:
                body += u"Gorjetas.........: [{0:>4}] R${1:>10}\n".format(qty_tip, self._str_float(value_tip))
                body += u"Venda s/ Gorjetas: [{0:>4}] R${1:>10}\n".format(qty_paid_orders, self._str_float(value_paid_orders + value_opened_orders))
            else:
                body += u"Taxas de Servico.: [{0:>4}] R${1:>10}\n".format(qty_tip, self._str_float(value_tip))
                body += u"Venda s/ Servico.: [{0:>4}] R${1:>10}\n".format(qty_paid_orders, self._str_float(value_paid_orders))
            print_donation_or_tip = True

        if print_donation_or_tip:
            body += DOUBLE_LINE_SEPARATOR

        body += self._write_payments(payments_by_brand, tef_card_brands)

        if report_type != self.TYPE_XML:
            body += u"Cancelamentos....: [{0:>4}] R${1:>10}\n".format(qty_total_voided, self._str_float(value_total_voided))
            body += u"  Pedidos........: [{0:>4}] R${1:>10}\n".format(qty_orders_voided, self._str_float(value_orders_voided))
            body += u"  Cupons.........: [{0:>4}] R${1:>10}\n".format(qty_coupons_voided, self._str_float(value_coupons_voided))
            if voided_lines is not None:
                body += u"Linhas Canceladas: [{0:>4}] R${1:>10}\n".format(qty_lines_voided, self._str_float(value_lines_voided))
            body += DOUBLE_LINE_SEPARATOR

        body = self._build_cash_in_and_cash_out(body, transfers)

        return body, additional_info

    def _build_cash_in_and_cash_out(self, body, transfers):
        x = {3: (0, 0), 4: (0, 0)}
        for transfer in transfers:
            if transfer.type not in [3, 4]:
                continue

            x[transfer.type] = (x[transfer.type][0] + 1, x[transfer.type][1] + transfer.amount)

        body += u"Sangrias.........: [{0:>4}] R${1:>10}\n".format(x[4][0], self._str_float(x[4][1]))
        body += u"Suprimentos......: [{0:>4}] R${1:>10}\n".format(x[3][0], self._str_float(x[3][1]))
        body += DOUBLE_LINE_SEPARATOR
        return body

    def _build_body_only_with_payments(self, payments_by_brand, tef_card_brands):
        body = u"Descricao          Qtd    Total\n"
        body += DOUBLE_LINE_SEPARATOR
        body += self._write_payments(payments_by_brand, tef_card_brands)
        return body

    @staticmethod
    def _str_float(float_num):
        return locale.format("%.2f", float_num, True)

    @staticmethod
    def fill_orders_sector(tables_by_sector, orders):
        # type: (Dict, List[Order]) -> List[Order]
        for order in orders:
            if order.table_id in tables_by_sector:
                source_table_id = tables_by_sector[order.table_id]["source_table_id"]
                if source_table_id is None:
                    order.sector = tables_by_sector[order.table_id]["sector"]
                else:
                    order.sector = tables_by_sector[source_table_id]["sector"]
        return orders

    @staticmethod
    def get_delivery_orders_values(delivery_paid_orders):
        delivery_orders_count = len([x for x in delivery_paid_orders])
        delivery_orders_value = sum([x.total for x in delivery_paid_orders])
        return delivery_orders_count, delivery_orders_value

    @staticmethod
    def get_opened_orders_values(pos_id, opened_orders):
        if not opened_orders:
            return 0, 0.0

        value_opened_orders = 0.0
        model = sysactions.get_model(pos_id)
        pos_ot = sysactions.get_posot(model)

        output = [opened_orders[i:i + 50] for i in range(0, len(opened_orders), 50)]
        for order_list in output:
            orders_pictures = pos_ot.orderPicture(pos_id, ";".join(order_list))
            orders_xml = eTree.XML(orders_pictures).findall('.//Order') if orders_pictures is not None else []
            for order in orders_xml:
                value_opened_orders += float(order.get("totalGross"))

        return len(opened_orders), value_opened_orders

    @staticmethod
    def _get_voided_order_values(voided_lines, voided_orders):
        qty_lines_voided, value_lines_voided = CashReport._get_voided_lines(voided_lines)
        qty_total_voided = len(voided_orders)
        value_total_voided = 0.0
        qty_orders_voided = 0
        value_orders_voided = 0.0
        qty_coupons_voided = 0
        value_coupons_voided = 0.0
        for order in voided_orders:
            if order.fiscal_canceled:
                qty_coupons_voided += 1
                value_coupons_voided += order.total
            else:
                qty_orders_voided += 1
                value_orders_voided += order.total
            value_total_voided += order.total

        return (
            qty_coupons_voided,
            qty_lines_voided,
            qty_orders_voided,
            qty_total_voided,
            value_coupons_voided,
            value_lines_voided,
            value_orders_voided,
            value_total_voided,
        )

    @staticmethod
    def _get_voided_lines(voided_lines):
        qty_lines_voided = int(sum(line.quantity for line in voided_lines) if voided_lines else 0)
        value_lines_voided = sum(line.price for line in voided_lines) if voided_lines else 0.0
        return qty_lines_voided, value_lines_voided

    @staticmethod
    def _get_initial_balance(is_table_service, transfers):
        initial_balance = 0.0
        if not is_table_service:
            for transfer in transfers:
                if transfer.type == 1:
                    initial_balance += transfer.amount
        return initial_balance

    def get_paid_order_values(self, paid_orders):
        qty_paid_orders = len(paid_orders)
        qty_discount = 0
        qty_donation = 0
        qty_tip = 0
        value_paid_orders = 0.0
        value_discount = 0.0
        value_donation = 0.0
        value_tip = 0.0
        for order in paid_orders:
            value_paid_orders += order.total
            qty_tip, value_tip = self._get_tip_values(order, qty_tip, value_tip)
            qty_donation, value_donation = self._get_donation_values(order, qty_donation, value_donation)
            qty_discount, value_discount = self._get_discount_values(order, qty_discount, value_discount)

        return qty_paid_orders, \
               qty_discount, \
               qty_donation, \
               qty_tip, \
               value_paid_orders, \
               value_discount, \
               value_donation, \
               value_tip

    @staticmethod
    def _get_discount_values(order, qty_discount, value_discount):
        if order.discount > 0:
            value_discount += order.discount
            qty_discount += 1
        return qty_discount, value_discount

    @staticmethod
    def _get_donation_values(order, qty_donation, value_donation):
        if order.donation > 0:
            value_donation += order.donation
            qty_donation += 1
        return qty_donation, value_donation

    @staticmethod
    def _get_tip_values(order, qty_tip, value_tip):
        if order.tip > 0:
            value_tip += order.tip
            qty_tip += 1
        return qty_tip, value_tip

    @staticmethod
    def get_grouped_paid_orders_by_sector(paid_orders):
        grouped_paid_orders_by_sector = {}
        for order in paid_orders:
            if order.sector:
                has_order_tip = (order.tip or 0)
                if order.sector not in grouped_paid_orders_by_sector:
                    grouped_paid_orders_by_sector[order.sector] = {"value": order.total + has_order_tip, "quantity": 1}
                else:
                    grouped_paid_orders_by_sector[order.sector]["value"] += order.total + has_order_tip
                    grouped_paid_orders_by_sector[order.sector]["quantity"] += 1

        return grouped_paid_orders_by_sector

    def build_table_info(self, average_info, grouped_paid_orders_by_sector, additional_info):
        """
        TM Balcao........: [XXXX] R$ 00.000,00
        TM Delivery......: [XXXX] R$ 00.000,00
        TM Mesa..........: [XXXX] R$ 00.000,00
          TM Cliente.....: [XXXX] R$ 00.000,00
          Giro Mesa......: hh:mm:ss

        ======================================
        :return: the report bytes encoded with UTF-8
        """

        value_paid = 0.0
        for sector in grouped_paid_orders_by_sector:
            value_paid += grouped_paid_orders_by_sector[sector]['value']

        seconds = (average_info.table_time_average_milliseconds / 1000) % 60
        seconds = int(seconds)
        minutes = (average_info.table_time_average_milliseconds / (1000 * 60)) % 60
        minutes = int(minutes)
        hours = (average_info.table_time_average_milliseconds / (1000 * 60 * 60)) % 24

        if average_info.total_tables != 0:
            tm = value_paid / average_info.total_tables
        else:
            tm = 0

        if average_info.total_customers != 0:
            tc = value_paid / average_info.total_customers
        else:
            tc = 0
        table_info = ""
        for item in additional_info:
            name = item["name"].decode("utf-8")
            value = self._str_float(item["value"])
            table_info += u"TM {0:.<14}: [{1:>4}] R${2:>10}\n".format(name, item["quantity"], value)
        table_info += u"TM Mesa..........: [{0:>4}] R${1:>10}\n".format(average_info.total_tables, self._str_float(tm))
        table_info += u"  TM Cliente.....: [{0:>4}] R${1:>10}\n".format(average_info.total_customers, self._str_float(tc))
        table_info += u"  Giro Mesa......:            {} \n".format("%02d:%02d:%02d" % (hours, minutes, seconds))
        table_info += DOUBLE_LINE_SEPARATOR

        return table_info

    def _write_payments(self, payments_by_brand, tef_card_brands):
        # type: (List[Tuple[unicode, int, float, int]], dict) -> unicode
        payments_text = u"Pagamentos.......: [{payment_qty:>4}] R${payment_value:>10}\n"
        payment_qty = 0
        payment_value = 0

        for card_brand in payments_by_brand:
            tender_id = [payments_by_brand[card_brand][x]['id'] for x in payments_by_brand[card_brand]][0]
            brand_quantity_sum = sum([payments_by_brand[card_brand][x]['quantity'] for x in payments_by_brand[card_brand]])
            brand_value_sum = sum([payments_by_brand[card_brand][x]['value'] for x in payments_by_brand[card_brand]])
            payment_qty += brand_quantity_sum
            payment_value += brand_value_sum

            if self.print_card_brands:
                payments_text += u"  {0:.<15}: [{1:>4}] R${2:>10}\n".format(
                    self.product_repository.get_tender_name(card_brand)[:15],
                    brand_quantity_sum,
                    self._str_float(brand_value_sum)
                )

                if tender_id in [TenderType.credit, TenderType.debit]:
                    for tef_brand, brand_tender_id in tef_card_brands:
                        if brand_tender_id == tender_id:
                            payments_text += u"    {0:.<13}: [{1:>4}] R${2:>10}\n".format(
                                tef_brand[:13],
                                tef_card_brands[(tef_brand, brand_tender_id)]["quantity"],
                                self._str_float(tef_card_brands[(tef_brand, brand_tender_id)]["value"])
                            )

                if len(payments_by_brand[card_brand]) == 1 and card_brand in payments_by_brand[card_brand]:
                    continue

                for sub_card_brand in payments_by_brand[card_brand]:
                    qty = payments_by_brand[card_brand][sub_card_brand]['quantity']
                    value = payments_by_brand[card_brand][sub_card_brand]['value']
                    payments_text += u"    {0:.<13}: [{1:>4}] R${2:>10}\n".format(
                        self.product_repository.get_tender_name(sub_card_brand)[:13],
                        qty,
                        self._str_float(value)
                    )

        payments_text += DOUBLE_LINE_SEPARATOR

        return payments_text.format(payment_qty=payment_qty, payment_value=self._str_float(payment_value))

    def _generate_cash_report_json(self, pos_id, report_pos, initial_date, end_date, operator_id, business_period, codigo_centro, close_time):
        # type: (int, Optional[unicode], datetime, datetime, unicode, unicode, unicode, datetime) -> str

        paid_orders = self.order_repository.get_paid_orders_by_business_period(initial_date, end_date, operator_id, report_pos)
        voided_orders = self.order_repository.get_voided_orders_by_business_period(initial_date, end_date, operator_id, report_pos)
        transfers = self.account_repository.get_transfers_by_business_period(initial_date, end_date, operator_id, report_pos)
        card_brands = self.fiscal_repository.get_card_brands(paid_orders)
        operators_qty = self.pos_ctrl_repository.get_operators_qty(pos_id, business_period)

        json_report = self._build_json(pos_id, paid_orders, voided_orders, transfers, operators_qty, business_period, codigo_centro, close_time, card_brands)

        return json_report.encode("utf-8")

    def _build_json(self, pos_id, paid_orders, voided_orders, transfers, operators_qty, business_period, codigo_centro, close_time, card_brands):
        # type: (int, List[Order], List[Order], List[Transfer], int, unicode, unicode, datetime, List[Tuple[unicode, int, float, int]]) -> unicode
        qty_paid_orders = len(paid_orders)
        value_paid_orders = 0.0
        total_tender_by_type_dict = {}  # type: Dict[int, TotalTender]
        total_tender = ''
        qty_donation = 0
        value_donation = 0.0
        for order in paid_orders:
            value_paid_orders += order.total
            qty_donation += 1 if order.donation > 0 else 0
            value_donation += order.donation
            self._calc_tenders_by_type(order, total_tender_by_type_dict)

        total_tender_by_type_list = []  # type: List[TotalTender]
        for tender_type_id in total_tender_by_type_dict:
            tender_type = total_tender_by_type_dict[tender_type_id]
            tender_type.name = total_tender.tender_name

            total_tender_by_type_list.append(tender_type)

        total_tender_by_type_list = sorted(total_tender_by_type_list, key=lambda x: x.tender_type)

        qty_voided_orders = len(voided_orders)
        value_voided_orders = 0.0
        for order in voided_orders:
            value_voided_orders += order.total

        initial_balance = 0.0
        value_sangria = 0.0
        qty_sangria = 0
        value_suprimento = 0.0
        qty_suprimento = 0
        for transfer in transfers:
            if transfer.type == 1:
                initial_balance += transfer.amount
            elif transfer.type == 2:
                value_sangria += transfer.amount
                qty_sangria += 1
            elif transfer.type == 3:
                value_suprimento += transfer.amount
                qty_suprimento += 1

        tender_types = {}

        for total_tender in total_tender_by_type_list:
            tender_types[total_tender.tender_name] = {'quantity': total_tender.quantity,
                                                      'amount': round(total_tender.value, 2),
                                                      'details': None}

            if total_tender.tender_type != 0:
                tender_types[total_tender.tender_name]['details'] = {}
                tender_total_qty = 0
                tender_total_value = 0

                for brand in card_brands:
                    if total_tender.tender_type == brand[1]:
                        tender_total_qty += brand[3]
                        tender_total_value += brand[2]
                        tender_types[total_tender.tender_name]['details'][brand[0]] = {'quantity': brand[3],
                                                                                       'amount': round(brand[2], 2)}
                tender_types[total_tender.tender_name]['quantity'] = tender_total_qty
                tender_types[total_tender.tender_name]['amount'] = round(tender_total_value, 2)

        response = {
            'branchNumber': str(self.store_id),
            'restaurantId': str(codigo_centro),
            'closeTime': close_time,
            'pdvNumber': pos_id,
            'businessPeriod': str(business_period),
            'initialBalance': round(initial_balance, 2),
            'operators': operators_qty,
            'ordersQuantity': qty_paid_orders + qty_voided_orders,
            'quantityPaidOrders': qty_paid_orders,
            'valuePaidOrders': round(value_paid_orders, 2),
            'quantityVoidedOrders': qty_voided_orders,
            'valueVoidedOrders': round(value_voided_orders, 2),
            'quantitySkim': qty_sangria,
            'valueSkim': round(value_sangria, 2),
            'quantityCashIn': qty_suprimento,
            'valueCashIn': round(value_suprimento, 2),
            'quantityDonation': qty_donation,
            'valueDonation': round(value_donation, 2),
            'tenderTypes': tender_types
        }

        return json.dumps(response)

    def generate_hourly_sale(self, report_pos, initial_date, end_date, operator):
        return self.order_repository.get_hourly_sale(report_pos, initial_date, end_date, operator)

    def generate_paid_order_cash_report_by_date(self, initial_date, end_date):
        return self.order_repository.get_paid_orders_by_real_date(initial_date, end_date, None, None)

    def generate_paid_orders_value_report(self, initial_date, end_date):
        # type: (datetime, datetime) -> unicode
        paid_orders = self.order_repository.get_paid_orders_by_business_period(initial_date, end_date, None, None)

        value_paid_orders = 0
        for order in paid_orders:
            value_paid_orders += order.total

        return unicode(value_paid_orders)

    def _manager_orders_id(self, orders):
        try:
            orders_id = self.fiscal_repository.get_order_id_by_nf_number(orders)
            for order in orders:
                if order.nf_number in orders_id:
                    order.order_id = orders_id[order.nf_number]
        except Exception as _:
            pass

        return orders



# -*- coding: utf-8 -*-

import base64
import json
from xml.etree import ElementTree as eTree
import sysactions
from common import PrinterCommon
from fiscalpersistence import Tender
from msgbus import MBEasyContext
from typing import (
    List,
    Optional,
    Dict,
)

from _AposA8PrintLine import AposA8PrintLine
from _AposA8Align import AposA8Align
from _AposA8CommandType import AposA8CommandType

from printer_wrapper import insert_space
from clichehelper import ClicheHelper
from _AposA8FiscalPrinter import AposA8FiscalPrinter
from _AposA8Helpers import (
    append_tax_data,
    APOS_DASH,
    split_lines,
    build_call_id,
)


class SatAposA8FiscalPrinter(AposA8FiscalPrinter):
    def __init__(
        self,
        mb_context,                     # type: MBEasyContext
        order_ref_ids,                  # type: list
        cliche_helper,                  # type: ClicheHelper
        customer_text_field,            # type: str
        printer_common,                 # type: PrinterCommon
        coupon_fiscal_parameters,       # type: Dict[str, bool]
        font_size_withdrawal_panel,     # type: str
        danfe_font_size_mobile,         # type: str
        order_ref_position,             # type: str
    ):
        # type: (...) -> None

        super(SatAposA8FiscalPrinter, self).__init__(
            mb_context=mb_context,
            order_ref_ids=order_ref_ids,
            cliche_helper=cliche_helper,
            customer_text_field=customer_text_field,
            coupon_fiscal_parameters=coupon_fiscal_parameters,
            font_size_withdrawal_panel=font_size_withdrawal_panel,
            danfe_font_size_mobile=danfe_font_size_mobile,
            order_ref_position=order_ref_position,
        )

        self.customer_text_field = customer_text_field
        self.printer_common = printer_common
        self.cliche_helper = cliche_helper
        self.font_size_withdrawal_panel = font_size_withdrawal_panel
        self.danfe_font_size_mobile = danfe_font_size_mobile

    def print_fiscal(
        self,
        pos_id,             # type: str
        order,              # type: eTree.Element
        tenders,            # type: list
        fiscal_request,     # type: eTree.Element
        fiscal_response,    # type: eTree.Element
        total_taxes,        # type: dict
        url_sefaz="",       # type: str
        is_reprint=False,   # type: Optional[bool]
    ):
        # type: (...) -> str

        xml = base64.b64decode(fiscal_request.text.strip())
        parsed_xml = eTree.XML(xml)
        n_c_fe = parsed_xml.find("infCFe/ide/nCFe").text
        model = sysactions.get_model(pos_id)

        report = self._fiscal_report(
            order=order,
            tenders=tenders,
            sat_fiscal_response=fiscal_response,
            n_c_fe=n_c_fe,
            model=model,
        )

        qr_code = fiscal_response.find("QRCode").text.strip()
        sat_key = fiscal_response.find("satkey").text.strip()
        report.append(AposA8PrintLine(AposA8Align.CENTER, sat_key, AposA8CommandType.PRINT_BAR_CODE))
        report.append(AposA8PrintLine(AposA8Align.CENTER, '', AposA8CommandType.PRINT_BAR_CODE))
        report.append(AposA8PrintLine(AposA8Align.CENTER, qr_code, AposA8CommandType.PRINT_QR_CODE))

        text = "Consulte o QRCode pelo Aplicativo De olho na nota disponivel na AppStore (Apple)"
        split_lines(
            report=report,
            text=text,
            font_size=self.danfe_font_size_mobile,
        )

        text = "e Play Store (Android)"
        report.append(self._get_text_line(text=text))

        report.append(APOS_DASH)
        append_tax_data(
            report=report,
            total_taxes=total_taxes,
            font_size=self.danfe_font_size_mobile,
        )

        return json.dumps(report, default=lambda x: x.__dict__)

    def print_cancel(self, posid, data):
        return

    def _cancel_report(self, sat_fiscal_response, nCFe, total):
        return

    def _fiscal_report(
        self,
        order,                  # type: eTree.Element
        tenders,                # type: List[Tender]
        sat_fiscal_response,    # type: eTree.Element
        n_c_fe,                 # type: str
        model=None,             # type: Optional[eTree]
    ):
        # type: (...) -> List[AposA8PrintLine]

        report = []

        if self.order_ref_position == "BEFORE_FISCAL_DATA":
            build_call_id(
                model=model,
                report=report,
                order=order,
                font_size_withdrawal_panel=self.font_size_withdrawal_panel,
                printer_common=self.printer_common,
                order_ref_ids=self.order_ref_ids,
                position="before",
            )

        self._append_cliche(report=report)

        self._append_header(
            report=report,
            n_c_fe=n_c_fe,
        )

        report.append(APOS_DASH)

        self._append_coupon_body(
            model=model,
            order=order,
            report=report,
            tenders=tenders,
            printer_common=self.printer_common,
        )

        report.append(APOS_DASH)

        if self.customer_text_field != "":
            for customer_text in self.customer_text_field.split('\\n'):
                report.append(self._get_text_line(text=customer_text.replace("\\n", "")))

            report.append(APOS_DASH)

        self._append_danfe_key(
            report=report,
            key=sat_fiscal_response.find("satkey").text.strip(),
        )
        report.append(APOS_DASH)

        cpf_cnpj = sat_fiscal_response.find("CustomerCPF").text
        cpf_text = "Consumidor: %s" % ("Nao informado" if not cpf_cnpj else cpf_cnpj)
        report.append(self._get_text_line(text=cpf_text))

        sat_number = "SAT No. %s" % sat_fiscal_response.find("satno").text.strip()
        report.append(self._get_text_line(text=sat_number))

        report.append(self._get_text_line(text=sat_fiscal_response.find("datetime").text.strip()))

        return report

    def _append_header(
        self,
        n_c_fe,     # type: str
        report,     # type: List[AposA8PrintLine]
    ):
        # type: (...) -> None

        report.append(
            self._get_text_line(
                text="EXTRATO No. %s" % n_c_fe,
            )
        )

        report.append(
            self._get_text_line(
                text="CUPOM FISCAL ELETRONICO - SAT",
            )
        )




from typing import Dict, Any
from abc import ABCMeta, abstractmethod
from pos_model import Order


class SatXmlPartBuilder(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode
        raise NotImplementedError()



from httpmessageprocessorutil.automapped import auto_map
from typing import Optional


@auto_map({"system_api_key": {"name": "X-Api-Key", "location": "header", "type": str},
           "pos_id": {"name": "posId", "type": str},
           "language": {"name": "language", "type": Optional[str]},
           "coupon_code": {"name": "couponCode", "type": str}})
class ApplyCouponRequest(object):
    def __init__(
        self,
        system_api_key,     # type: str
        pos_id,             # type: str
        coupon_code,        # type: str
        language,           # type: str
    ):
        # type: (...) -> None

        self.system_api_key = system_api_key
        self.pos_id = pos_id
        self.language = language
        self.coupon_code = coupon_code



from typing import Any

from messagebus import (
    AckMessage,
    Event,
    Message,
    MessageBus,
    NakMessage,
)
from messageprocessorutil import UuidMessageProcessor

from services.interactor import UpdateAlertsInteractor


class UpdateAlertsProcessor(UuidMessageProcessor):
    def __init__(
        self,
        update_alerts_interactor,  # type: UpdateAlertsInteractor
    ):
        # type: (...) -> None

        super(UpdateAlertsProcessor, self).__init__()
        self.update_alerts_interactor = update_alerts_interactor

    def parse_data(
        self,
        data,           # type: str
        event_type,     # type: str
    ):
        # type: (...) -> None

        return None

    def call_business(
        self,
        request,    # type: None
    ):
        # type: (...) -> str

        return self.update_alerts_interactor.process_request()

    def format_response(
        self,
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Event
        request,        # type: None
        result,         # type: None
    ):
        # type: (...) -> None

        message_bus.reply_message(
            message=message,
            reply_message=AckMessage(data=result),
        )

    def format_exception(
        self,
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Event
        input_obj,      # type: None
        exception,      # type: Exception
    ):
        # type: (...) -> None

        message_bus.reply_message(
            message=message,
            reply_message=NakMessage(data=exception.message),
        )

    def format_parse_exception(
        self,
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Event
        data,           # type: Any
        exception,      # type: Exception
    ):
        # type: (...) -> None

        message_bus.reply_message(
            message=message,
            reply_message=NakMessage(data=exception.message),
        )



import logging
import sys

from application.model import AuditEvent
from application.repository import MessageBusRepository
from cfgtools import read
from helper import config_logger
from syserrors import SE_CFGLOAD, SE_NOTFOUND


class Configurator(object):

    def __init__(self, loader_path):
        # type: (str) -> None

        self.logger = None
        self.loader_path = loader_path
        self.cfg_reader = read(loader_path)
        self.cfg_group = None
        self.service_name = None
        self.required_services = "AuditLogger|StoreWideConfig"
        self.service_type = "DataPump"
        order_picture_url = "pump/sale/order-picture"
        self.views = {
            AuditEvent.PAID: order_picture_url,
            AuditEvent.VOID_ORDER: order_picture_url,
            AuditEvent.ORDER_REPROCESS: order_picture_url,
            AuditEvent.CLEAR_TENDERS: order_picture_url,
            AuditEvent.OPERATOR_LOGOUT: "pump/cash/close-drawer-v2",
            AuditEvent.PRODUCTION_ORDER: "pump/sale/production-order",
            AuditEvent.TABLE_CLOSED: "pump/sale/tablePicture",
            AuditEvent.CLOSE_STORE_DAY: "pump/????",
            AuditEvent.OPEN_STORE_DAY: "pump/?????",
        }
        self.pump_status_url = "pump/sale/pumpStatus"

        self.store_id = None
        self.store_division = None
        self.server_url = None
        self.api_key = None
        self.retry_wait = None
        self.fiscal_xml = None

    def configure_service_name(self):
        # type: () -> None

        self.service_name = self.cfg_group = str(self.cfg_reader.find_value("Identification.ComponentName"))

    def configure_logger(self):
        # type: () -> None

        config_logger(self.loader_path, self.service_name)
        self.logger = logging.getLogger(self.service_name)

    def read_configurations(self, pos_repository):
        # type: (MessageBusRepository) -> None

        try:
            self.logger.info("[START] Reading configurations")

            group = self.service_name + ".{}"

            self.store_id = pos_repository.read_sw_config("Store.Id")
            self.store_division = pos_repository.read_sw_config("Store.Division")

            self._read_api_key(pos_repository, group)
            self._read_server_url(pos_repository, group)

            self.retry_wait = int(self.cfg_reader.find_value(group.format("retryWait"), "10000"))
            self.fiscal_xml = self.cfg_reader.find_value(group.format("fiscalXml"), "true").lower() == "true"

            self.logger.info("[END] Reading configurations")
        except Exception as _:
            self.logger.exception("[{}] Error reading configuration file".format(self.service_name))
            sys.exit(SE_CFGLOAD)

    def _read_api_key(self, pos_repository, group):
        # type: (MessageBusRepository, str) -> None

        self.api_key = self.cfg_reader.find_value(group.format("OverwrittenApiKey"))
        if not self.api_key:
            self.api_key = pos_repository.read_sw_config("BackOffice.ApiKey")

    def _read_server_url(self, pos_repository, group):
        # type: (MessageBusRepository, str) -> None

        self.server_url = self.cfg_reader.find_value(group.format("OverwrittenServerURL"))
        if not self.server_url:
            self.server_url = pos_repository.read_sw_config("BackOffice.ServerURL")
        if not self.server_url:
            self.logger.error("[{}] Configuration error: server_url is undefined".format(self.service_name))
            sys.exit(SE_NOTFOUND)
        if not self.server_url.endswith("/"):
            self.server_url += "/"

        self.logger.info("POST URL: {}".format(self.server_url))



# -*- coding: utf-8 -*-

import pyscripts
import sysactions
from bustoken import (
    TK_SLCTRL_PMGR_GET_NAVIGATION,
)
from msgbus import (
    FM_PARAM,
    TK_SYS_ACK,
    MBEasyContext,
)
from posot import OrderTakerException

mb_context = pyscripts.mbcontext  # type: MBEasyContext


@sysactions.action
def get_navigation_data(
    pos_id,     # type: str
):
    # type: (...) -> str

    message_response = mb_context.MB_EasySendMessage(
        dest_name='ORDERMGR{}'.format(
            pos_id
        ),
        token=TK_SLCTRL_PMGR_GET_NAVIGATION,
        format=FM_PARAM,
        data=pos_id,
    )

    if message_response.token != TK_SYS_ACK:
        raise OrderTakerException(
            code=-1,
            descr=message_response.data,
        )
    return message_response.data



from application.service import BenefitService
from application.model import Configurations
from typing import List


class RemoveBenefitInteractor(object):

    def __init__(self, configs, benefit_service):
        # type: (Configurations, BenefitService) -> None

        self.configs = configs
        self.benefit_service = benefit_service

    def remove(self, pos_id, benefit_keys):
        # type: (int, List) -> None

        static_uow = self.benefit_service.get_static_unit_of_work(pos_id=pos_id)
        static_uow.pos_ot.blkopnotify = True

        self.benefit_service.remove_benefits(
            pos_id=pos_id,
            benefit_keys=benefit_keys,
            static_uow=static_uow,
        )

        self.benefit_service.apply_pending_benefits(
            pos_id=pos_id,
            static_uow=static_uow,
        )

        self.benefit_service.send_order_update_event(static_uow=static_uow)



from typing import List

from ._RequesterType import RequesterType


class DeviceConfig(object):
    def __init__(
        self,
        requester_type,    # type: RequesterType
        requester_id,      # type: int
        default_device,    # type: str
        fallback_devices,  # type: List[str]
    ):
        # type: (...) -> None

        self.requester_type = requester_type
        self.requester_id = requester_id
        self.default_device = default_device
        self.fallback_devices = fallback_devices



# Copyright Jonathan Hartley 2013. BSD 3-Clause license, see LICENSE file.
import sys
from unittest import TestCase, main, skipUnless

try:
    from unittest.mock import Mock, patch
except ImportError:
    from mock import Mock, patch

from ..winterm import WinColor, WinStyle, WinTerm


class WinTermTest(TestCase):

    @patch('colorama.winterm.win32')
    def testInit(self, mockWin32):
        mockAttr = Mock()
        mockAttr.wAttributes = 7 + 6 * 16 + 8
        mockWin32.GetConsoleScreenBufferInfo.return_value = mockAttr
        term = WinTerm()
        self.assertEqual(term._fore, 7)
        self.assertEqual(term._back, 6)
        self.assertEqual(term._style, 8)

    @skipUnless(sys.platform.startswith("win"), "requires Windows")
    def testGetAttrs(self):
        term = WinTerm()

        term._fore = 0
        term._back = 0
        term._style = 0
        self.assertEqual(term.get_attrs(), 0)

        term._fore = WinColor.YELLOW
        self.assertEqual(term.get_attrs(), WinColor.YELLOW)

        term._back = WinColor.MAGENTA
        self.assertEqual(
            term.get_attrs(),
            WinColor.YELLOW + WinColor.MAGENTA * 16)

        term._style = WinStyle.BRIGHT
        self.assertEqual(
            term.get_attrs(),
            WinColor.YELLOW + WinColor.MAGENTA * 16 + WinStyle.BRIGHT)

    @patch('colorama.winterm.win32')
    def testResetAll(self, mockWin32):
        mockAttr = Mock()
        mockAttr.wAttributes = 1 + 2 * 16 + 8
        mockWin32.GetConsoleScreenBufferInfo.return_value = mockAttr
        term = WinTerm()

        term.set_console = Mock()
        term._fore = -1
        term._back = -1
        term._style = -1

        term.reset_all()

        self.assertEqual(term._fore, 1)
        self.assertEqual(term._back, 2)
        self.assertEqual(term._style, 8)
        self.assertEqual(term.set_console.called, True)

    @skipUnless(sys.platform.startswith("win"), "requires Windows")
    def testFore(self):
        term = WinTerm()
        term.set_console = Mock()
        term._fore = 0

        term.fore(5)

        self.assertEqual(term._fore, 5)
        self.assertEqual(term.set_console.called, True)

    @skipUnless(sys.platform.startswith("win"), "requires Windows")
    def testBack(self):
        term = WinTerm()
        term.set_console = Mock()
        term._back = 0

        term.back(5)

        self.assertEqual(term._back, 5)
        self.assertEqual(term.set_console.called, True)

    @skipUnless(sys.platform.startswith("win"), "requires Windows")
    def testStyle(self):
        term = WinTerm()
        term.set_console = Mock()
        term._style = 0

        term.style(22)

        self.assertEqual(term._style, 22)
        self.assertEqual(term.set_console.called, True)

    @patch('colorama.winterm.win32')
    def testSetConsole(self, mockWin32):
        mockAttr = Mock()
        mockAttr.wAttributes = 0
        mockWin32.GetConsoleScreenBufferInfo.return_value = mockAttr
        term = WinTerm()
        term.windll = Mock()

        term.set_console()

        self.assertEqual(
            mockWin32.SetConsoleTextAttribute.call_args,
            ((mockWin32.STDOUT, term.get_attrs()), {})
        )

    @patch('colorama.winterm.win32')
    def testSetConsoleOnStderr(self, mockWin32):
        mockAttr = Mock()
        mockAttr.wAttributes = 0
        mockWin32.GetConsoleScreenBufferInfo.return_value = mockAttr
        term = WinTerm()
        term.windll = Mock()

        term.set_console(on_stderr=True)

        self.assertEqual(
            mockWin32.SetConsoleTextAttribute.call_args,
            ((mockWin32.STDERR, term.get_attrs()), {})
        )


if __name__ == '__main__':
    main()



from helper import ExtendedEnum


class ApplyCouponErrorCode(ExtendedEnum):
    COUPON_NOT_FOUND = "The voucher id was not found"
    CUSTOM_ERROR = "The voucher is not available based on custom partner rules"
    COUPON_LOCKED = "The voucher is being used in another pos"
    COUPON_ALREADY_USED = "The voucher was already used"
    NDISCOUNT_ERROR = "The voucher could no be used because the conditions where not validated by the nDiscount"



"""
Middleware
==========

A WSGI middleware is a WSGI application that wraps another application
in order to observe or change its behavior. Werkzeug provides some
middleware for common use cases.

.. toctree::
    :maxdepth: 1

    proxy_fix
    shared_data
    dispatcher
    http_proxy
    lint
    profiler

The :doc:`interactive debugger </debug>` is also a middleware that can
be applied manually, although it is typically used automatically with
the :doc:`development server </serving>`.

:copyright: 2007 Pallets
:license: BSD-3-Clause
"""



from multiprocessing import Condition


class MobileServiceFinder(object):
    def __init__(self):
        self.available_sitef_lock = Condition()

    def lock_sitef(self):
        self.available_sitef_lock.acquire()

    def release_sitef(self):
        self.available_sitef_lock.release()



from basereport import I18n, I18nReporter

from ._MwI18nRepository import MwI18nRepository


class ReporterFactory(object):
    @staticmethod
    def build_reporter(pos_id, generator, presenter):
        i18n_repository = MwI18nRepository(pos_id)
        i18n = I18n(i18n_repository)
        return I18nReporter(generator, presenter, i18n)



from ._CommonProcessor import _CommonProcessor


class CancelPaymentProcessor(_CommonProcessor):
    def parse_data(self, data, event_type):
        return None

    def call_business(self, _):
        processor = self.payment_router.get_payment_processor()
        if processor:
            return processor.cancel_payment()
